from ctypes import c_double
import os
import multiprocessing
from queue import Queue, Empty
import shutil
import tempfile
from textwrap import dedent
import unittest
from unittest.mock import MagicMock

from green.process import ProcessLogger, poolRunner
from green import process


class TestProcessLogger(unittest.TestCase):
    def test_callThrough(self):
        """
        Calls are passed through to the wrapped callable
        """
        message = "some message"

        def func():
            return message

        l = ProcessLogger(func)
        self.assertEqual(l(), message)

    def test_exception(self):
        """
        A raised exception gets re-raised
        """
        saved_get_logger = process.multiprocessing.get_logger
        mock_logger = MagicMock()

        def addHandler(ignore):
            mock_logger.handlers = [MagicMock()]

        mock_logger.addHandler = addHandler
        mock_logger.handlers = False
        mock_get_logger = MagicMock()
        mock_get_logger.return_value = mock_logger
        process.multiprocessing.get_logger = mock_get_logger
        self.addCleanup(
            setattr, process.multiprocessing, "get_logger", saved_get_logger
        )

        def func():
            raise AttributeError

        l = ProcessLogger(func)
        self.assertRaises(AttributeError, l)
        mock_get_logger.assert_any_call()


class TestPoolRunner(unittest.TestCase):
    # Setup
    @classmethod
    def setUpClass(cls):
        cls.startdir = os.getcwd()
        cls.container_dir = tempfile.mkdtemp()

    @classmethod
    def tearDownClass(cls):
        if os.getcwd() != cls.startdir:
            os.chdir(cls.startdir)
        cls.startdir = None
        shutil.rmtree(cls.container_dir)

    def setUp(self):
        os.chdir(self.container_dir)
        self.tmpdir = tempfile.mkdtemp(dir=self.container_dir)

    def tearDown(self):
        os.chdir(self.container_dir)
        shutil.rmtree(self.tmpdir)

    # Tests
    def test_normalRun(self):
        """
        Runs normally
        """
        saved_coverage = process.coverage
        process.coverage = MagicMock()
        self.addCleanup(setattr, process, "coverage", saved_coverage)
        # Parent directory setup
        os.chdir(self.tmpdir)
        sub_tmpdir = tempfile.mkdtemp(dir=self.tmpdir)
        basename = os.path.basename(sub_tmpdir)
        # Child setup
        fh = open(os.path.join(basename, "__init__.py"), "w")
        fh.write("\n")
        fh.close()
        fh = open(os.path.join(basename, "test_pool_runner_dotted.py"), "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testPass(self):
                    pass
            """
            )
        )
        fh.close()
        module_name = basename + ".test_pool_runner_dotted.A.testPass"
        result = Queue()
        poolRunner(module_name, result, 1)
        result.get()
        self.assertEqual(len(result.get().passing), 1)

    def test_SyntaxErrorInUnitTest(self):
        """
        SyntaxError gets reported as an error loading the unit test
        """
        saved_coverage = process.coverage
        process.coverage = MagicMock()
        self.addCleanup(setattr, process, "coverage", saved_coverage)
        # Parent directory setup
        os.chdir(self.tmpdir)
        sub_tmpdir = tempfile.mkdtemp(dir=self.tmpdir)
        basename = os.path.basename(sub_tmpdir)
        # Child setup
        fh = open(os.path.join(basename, "__init__.py"), "w")
        fh.write("\n")
        fh.close()
        fh = open(os.path.join(basename, "test_pool_syntax_error.py"), "w")
        fh.write("aoeu")
        fh.close()
        result = Queue()
        poolRunner(basename, result, 1)
        result.get()
        self.assertEqual(len(result.get().errors), 1)

    def test_error(self):
        """
        Exception raised running unit test is reported as an error
        """
        # Parent directory setup
        os.chdir(self.tmpdir)
        sub_tmpdir = tempfile.mkdtemp(dir=self.tmpdir)
        basename = os.path.basename(sub_tmpdir)
        # Child setup
        fh = open(os.path.join(basename, "__init__.py"), "w")
        fh.write("\n")
        fh.close()
        fh = open(os.path.join(basename, "test_pool_runner_dotted_fail.py"), "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testError(self):
                    raise AttributeError
            """
            )
        )
        fh.close()
        module_name = basename + ".test_pool_runner_dotted_fail.A.testError"
        result = Queue()
        poolRunner(module_name, result)
        result.get()
        self.assertEqual(len(result.get().errors), 1)

    def test_bad_attr(self):
        """
        Accessing a bad attribute is only reported once (see #150)
        """
        # Parent directory setup
        os.chdir(self.tmpdir)
        sub_tmpdir = tempfile.mkdtemp(dir=self.tmpdir)
        basename = os.path.basename(sub_tmpdir)
        # Child setup
        fh = open(os.path.join(basename, "__init__.py"), "w")
        fh.write("\n")
        fh.close()
        fh = open(os.path.join(basename, "test_pool_runner_bad_attr.py"), "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testBadAttr(self):
                    "".garbage
            """
            )
        )
        fh.close()
        module_name = basename + ".test_pool_runner_bad_attr.A.testBadAttr"
        result = Queue()
        poolRunner(module_name, result)
        result.get_nowait()  # should get the target name
        result.get_nowait()  # should get the result
        result.get_nowait()  # should get None
        # should raise Empty unless the extra result bug is present
        self.assertRaises(Empty, result.get_nowait)

    def test_process(self):
        """
                Avoid FileNotFoundError when using a multiprocessing.Value, fixes #154.
                This test never fails, we have to watch the outer stderr to see if we get output like this:

          File "/usr/local/Cellar/python3/3.5.2_3/Frameworks/Python.framework/Versions/3.5/lib/python3.5/multiprocessing/util.py", line 254, in _run_finalizers
            finalizer()
          File "/usr/local/Cellar/python3/3.5.2_3/Frameworks/Python.framework/Versions/3.5/lib/python3.5/multiprocessing/util.py", line 186, in __call__
            res = self._callback(*self._args, **self._kwargs)
          File "/usr/local/Cellar/python3/3.5.2_3/Frameworks/Python.framework/Versions/3.5/lib/python3.5/shutil.py", line 465, in rmtree
            onerror(os.lstat, path, sys.exc_info())
          File "/usr/local/Cellar/python3/3.5.2_3/Frameworks/Python.framework/Versions/3.5/lib/python3.5/shutil.py", line 463, in rmtree
            orig_st = os.lstat(path)
        FileNotFoundError: [Errno 2] No such file or directory: '/var/folders/8y/cgqfhxyn2fz3r8n627_6dm_m0000gn/T/tmpp3fobx6i/pymp-8hpbali9'

                Newer versions of Python want to do their own cleanup, so let them.
        """
        val = multiprocessing.Value(c_double, 0)
        # The error happens when something tries to clean up a sub-temporary
        # directory that they assume will always be there to be cleaned up.



from __future__ import annotations  # pragma: no cover

import pathlib  # pragma: no cover
import sys  # pragma: no cover

import coverage  # pragma: no cover

__version__: str = (
    (pathlib.Path(__file__).parent / "VERSION").read_text(encoding="utf-8").strip()
)  # pragma: no cover


def pretty_version() -> str:  # pragma: no cover
    python_version = ".".join(str(x) for x in sys.version_info[0:3])
    return (
        f"Green {__version__}, Coverage {coverage.__version__}, Python {python_version}"
    )



import argparse
import contextlib
import os
import sys
import unittest
from configparser import ConfigParser
from unittest.mock import patch, MagicMock, call

from setuptools.dist import Distribution

from green import command
from green.config import StoreOpt


class TestCommand(unittest.TestCase):
    @contextlib.contextmanager
    def environ(self, setup_cfg=None, *args, **variables):
        args = ["green"] + list(args)

        if setup_cfg is not None:
            parser = ConfigParser()
            parser.add_section("green")
            for k, v in setup_cfg.items():
                parser.set("green", k, str(v))
            with open("setup.cfg", "w") as f:
                parser.write(f)

        yield Distribution(
            {"script_name": "setup.py", "script_args": args or ["green"]}
        )

        # finally:
        if os.path.isfile("setup.cfg"):
            os.remove("setup.cfg")

    def test_get_user_options(self):
        # Check the user options contain some of the
        # actual command line options
        options = command.get_user_options()

        self.assertIn(
            ("options", None, "Output all options.  Used by bash- and zsh-completion."),
            options,
        )

        self.assertIn(
            ("file-pattern=", "p", "Pattern to match test files. Default is test*.py"),
            options,
        )

    def test_get_user_options_setup_py(self):
        """
        When get_user_options() is called by 'python setup.py --help-commands',
        it returns [] early and doesn't crash.
        """
        sys.argv.append("--help-commands")
        self.addCleanup(lambda: sys.argv.pop())

        self.assertEqual(command.get_user_options(), [])

    @patch("green.command.parseArguments")
    def test_get_user_options_dynamic(self, parseArguments):
        # Check the user options are generated after
        # the command line options created in green.cmdline.parseArguments
        store_opt = StoreOpt()
        argparser = argparse.ArgumentParser()
        store_opt(argparser.add_argument("-s", "--something", help="Something"))
        store_opt(argparser.add_argument("--else", help="Else"))
        store_opt(
            argparser.add_argument("-a", "--again", action="store_true", help="Again")
        )

        args = argparser.parse_args([])
        args.parser = argparser
        args.store_opt = store_opt
        parseArguments.return_value = args

        options = command.get_user_options()

        self.assertEqual(
            options,
            [
                ("something=", "s", "Something"),
                ("else=", None, "Else"),
                ("again", "a", "Again"),
            ],
        )

    def test_initialize_options(self):
        d = Distribution({"script_name": "setup.py", "script_args": ["green"]})

        cmd = command.green(d)
        for attr in ["completion_file", "clear_omit", "debug", "processes"]:
            self.assertTrue(hasattr(cmd, attr), attr)

    @patch("green.command.main", return_value=0)
    def test_run(self, main):
        d = Distribution(
            {
                "script_name": "setup.py",
                "script_args": ["green"],
                "test_suite": "green.test.test_version",
            }
        )

        cmd = command.green(d)
        cmd.run()
        main.assert_called_once_with(["green.test.test_version"])

    @patch("green.command.main", return_value=125)
    def test_run_exits(self, main):
        d = Distribution({"script_name": "setup.py", "script_args": ["green"]})

        cmd = command.green(d)
        with self.assertRaises(SystemExit) as se:
            cmd.run()
        self.assertEqual(se.exception.code, 125)

    @patch("green.command.main", return_value=0)
    def test_requires(self, main):
        d = Distribution(
            {
                "script_name": "setup.py",
                "script_args": ["green"],
                "install_requires": ["six"],
                "tests_require": ["mock", "unittest2"],
            }
        )
        d.fetch_build_eggs = MagicMock()
        cmd = command.green(d)
        cmd.run()

        d.fetch_build_eggs.assert_has_calls(
            [
                call(["six"]),
                call(["mock", "unittest2"]),
            ]
        )



"""Registers the green command with setuptools."""

from __future__ import annotations

import functools
import sys
from typing import TYPE_CHECKING

from setuptools import Command

from green.config import parseArguments
from green.cmdline import main

if TYPE_CHECKING:
    from argparse import Action


def get_user_options() -> list[tuple[str, str | None, str | None]]:
    # When running "python setup.py --help-commands", setup.py will call this
    # function -- but green isn't actually being called.
    if "--help-commands" in sys.argv:
        return []

    args = parseArguments()
    options: list[tuple[str, str | None, str | None]] = []

    action: Action
    for action in args.store_opt.actions:
        names = [name.lstrip("-") for name in action.option_strings]
        short_name: str | None
        if len(names) == 1:
            full_name = names[0]
            short_name = None
        else:
            # TODO: We might want to pick the longer of the two for full_name.
            full_name = names[1]
            short_name = names[0]
        if not action.const:
            full_name += "="
        options.append((full_name, short_name, action.help))

    return options


class green(Command):
    command_name = "green"
    description = "Run unit tests using green"

    @functools.cached_property
    def user_options(self) -> list[tuple[str, str | None, str | None]]:
        return get_user_options()

    def initialize_options(self) -> None:
        for name, _, _ in self.user_options:
            setattr(self, name.replace("-", "_").rstrip("="), None)

    def finalize_options(self) -> None:
        pass

    def run(self) -> None:
        self.ensure_finalized()

        if self.distribution.install_requires:
            self.distribution.fetch_build_eggs(self.distribution.install_requires)
        if self.distribution.tests_require:
            self.distribution.fetch_build_eggs(self.distribution.tests_require)

        script_args = self.distribution.script_args[1:]
        if self.distribution.test_suite is not None:
            script_args.append(self.distribution.test_suite)

        error_code = main(script_args)
        if error_code:
            sys.exit(error_code)



import copy
import multiprocessing
import os
import pathlib
import shutil
import subprocess
import sys
import tempfile
from textwrap import dedent
import unittest


class TestFinalizer(unittest.TestCase):
    def setUp(self) -> None:
        self.tmpdir = pathlib.Path(tempfile.mkdtemp())

    def tearDown(self) -> None:
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    def test_finalizer(self) -> None:
        """
        Test that the finalizer works on Python 3.8+.
        """
        sub_tmpdir = pathlib.Path(tempfile.mkdtemp(dir=self.tmpdir))
        for i in range(multiprocessing.cpu_count() * 2):
            finalizer_path = sub_tmpdir / f"test_finalizer.py"
            finalizer_path.write_text(
                dedent(
                    f"""
                import unittest
                class Pass{i}(unittest.TestCase):
                    def test_pass{i}(self):
                        pass
                def msg():
                    print("finalizer worked")
                """
                )
            )
        args = [
            sys.executable,
            "-m",
            "green.cmdline",
            "--finalizer=test_finalizer.msg",
            "--maxtasksperchild=1",
        ]
        pythonpath = str(pathlib.Path(__file__).parent.parent.parent)

        env = copy.deepcopy(os.environ)
        env["PYTHONPATH"] = pythonpath

        output = subprocess.run(
            args,
            cwd=str(sub_tmpdir),
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=env,
            timeout=10,
            encoding="utf-8",
            check=True,
        ).stdout
        self.assertIn("finalizer worked", output)



from __future__ import annotations

import configparser
import copy
import pathlib
from io import StringIO
import os
import shutil
import tempfile
import unittest
from typing import Sequence

from green import config
from green.output import GreenStream


class ParseArguments(unittest.TestCase):
    def test_target(self):
        """
        The specified target gets parsed
        """
        config.sys.argv = ["", "target1", "target2"]
        args = config.parseArguments()
        self.assertEqual(args.targets, ["target1", "target2"])

    def test_absent(self):
        """
        Arguments not specified on the command-line are not present in the args
        object.
        """
        config.sys.argv = ["", "--debug"]
        args = config.parseArguments()
        self.assertEqual(getattr(args, "debug", "not there"), True)
        self.assertEqual(getattr(args, "verbose", "not there"), "not there")
        self.assertEqual(getattr(args, "targets", "not there"), "not there")
        self.assertEqual(getattr(args, "file_pattern", "not there"), "not there")


class ModifiedEnvironment:
    """
    I am a context manager that sets up environment variables for a test case.
    """

    def __init__(self, **kwargs: str | None) -> None:
        self.prev = {}
        self.excur = kwargs
        for k in kwargs:
            self.prev[k] = os.getenv(k)

    def __enter__(self) -> None:
        self.update_environment(self.excur)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.update_environment(self.prev)

    def update_environment(self, env: dict[str, str | None]) -> None:
        for key, value in env.items():
            if value is None:
                if key in os.environ:
                    del os.environ[key]
            else:
                os.environ[key] = value


class ConfigBase(unittest.TestCase):
    """
    I am an abstract base class that creates and destroys configuration files
    in a temporary directory with known values attached to self.
    """

    def _write_file(self, path: pathlib.Path, lines: Sequence[str]) -> None:
        path.write_text("\n".join(lines) + "\n")

    def setUp(self):
        self.tmpd = tmpd = pathlib.Path(tempfile.mkdtemp())
        self.addCleanup(shutil.rmtree, tmpd)
        # Set CWD to known empty directory so we don't pick up some other .green
        # file from the CWD tests are actually run from.
        self.addCleanup(os.chdir, pathlib.Path.cwd())
        cwd_dir = tmpd / "cwd"
        cwd_dir.mkdir(exist_ok=True, parents=True)
        os.chdir(cwd_dir)
        # This represents the $HOME config file, and doubles for the current
        # working directory config file if we set CWD to self.tmpd
        self.default_filename = tmpd / ".green"
        self.default_logging = False
        self.default_version = False
        self.default_failfast = True
        self.default_termcolor = True
        self._write_file(
            self.default_filename,
            [
                "# this is a test config file for green",
                f"logging = {self.default_logging}",
                f"version = {self.default_version}",
                f"omit-patterns = {self.default_filename}",
                f"failfast = {self.default_failfast}",
                f"termcolor = {self.default_termcolor}",
            ],
        )
        self.env_filename = tmpd / "green.env"
        self.env_logging = True
        self.env_no_skip_report = False
        self._write_file(
            self.env_filename,
            [
                "# this is a test config file for green",
                f"logging = {self.env_logging}",
                f"omit-patterns = {self.env_filename}",
                f"no-skip-report = {self.env_no_skip_report}",
            ],
        )
        self.cmd_filename = self.tmpd / "green.cmd"
        self.cmd_logging = False
        self.cmd_run_coverage = False
        self._write_file(
            self.cmd_filename,
            [
                "# this is a test config file for green",
                f"logging = {self.cmd_logging}",
                f"omit-patterns = {self.cmd_filename}",
                f"run-coverage = {self.cmd_run_coverage}",
            ],
        )
        self.setup_filename = cwd_dir / "setup.cfg"
        self.setup_failfast = False
        self.setup_verbose = 3
        self._write_file(
            self.setup_filename,
            [
                "[green]",
                f"failfast = {self.setup_failfast}",
                f"verbose = {self.setup_verbose}",
            ],
        )


class TestConfig(ConfigBase):
    """
    All variations of config file parsing works as expected.
    """

    def test_cmd_env_nodef_nosetup(self):
        """
        Setup: --config on cmd, $GREEN_CONFIG is set, $HOME/.green does not
            exist, setup.cfg does not exist
        Result: load --config
        """
        self.default_filename.unlink(missing_ok=True)
        self.setup_filename.unlink(missing_ok=True)
        with ModifiedEnvironment(
            GREEN_CONFIG=str(self.env_filename), HOME=str(self.tmpd)
        ):
            cfg = config.getConfig(self.cmd_filename)
            self.assertEqual(["green"], cfg.sections())
            self.assertEqual(str(self.cmd_filename), cfg.get("green", "omit-patterns"))
            self.assertEqual(
                self.cmd_run_coverage, cfg.getboolean("green", "run-coverage")
            )
            self.assertEqual(self.cmd_logging, cfg.getboolean("green", "logging"))
            self.assertEqual(
                self.env_no_skip_report, cfg.getboolean("green", "no-skip-report")
            )
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "version"
            )
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "verbose"
            )

    def test_cmd_noenv_def_nosetup(self):
        """
        Setup: --config on cmd, $GREEN_CONFIG unset, $HOME/.green exists,
            setup.cfg does not exist
        Result: load --config
        """
        os.unlink(self.env_filename)
        os.remove(self.setup_filename)
        with ModifiedEnvironment(GREEN_CONFIG=None, HOME=str(self.tmpd)):
            cfg = config.getConfig(self.cmd_filename)
            self.assertEqual(["green"], cfg.sections())
            self.assertEqual(str(self.cmd_filename), cfg.get("green", "omit-patterns"))
            self.assertEqual(
                self.cmd_run_coverage, cfg.getboolean("green", "run-coverage")
            )
            self.assertEqual(self.cmd_logging, cfg.getboolean("green", "logging"))
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "no-skip-report"
            )
            self.assertEqual(self.default_version, cfg.getboolean("green", "version"))
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "verbose"
            )

    def test_cmd_noenv_nodef_nosetup(self):
        """
        Setup: --config on cmd, $GREEN_CONFIG unset, $HOME/.green does not
            exist, setup.cfg does not exist
        Result: load --config
        """
        os.unlink(self.env_filename)
        os.unlink(self.default_filename)
        os.remove(self.setup_filename)
        with ModifiedEnvironment(GREEN_CONFIG=None, HOME=str(self.tmpd)):
            cfg = config.getConfig(self.cmd_filename)
            self.assertEqual(["green"], cfg.sections())
            self.assertEqual(str(self.cmd_filename), cfg.get("green", "omit-patterns"))
            self.assertEqual(
                self.cmd_run_coverage, cfg.getboolean("green", "run-coverage")
            )
            self.assertEqual(self.cmd_logging, cfg.getboolean("green", "logging"))
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "no-skip-report"
            )
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "version"
            )
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "verbose"
            )

    def test_nocmd_env_cwd(self):
        """
        Setup: no --config option, $GREEN_CONFIG is set, .green in local dir,
        Result: load $GREEN_CONFIG
        """
        os.chdir(self.tmpd)  # setUp is already set to restore us to our pre-testing cwd
        os.unlink(self.cmd_filename)
        os.remove(self.setup_filename)
        with ModifiedEnvironment(
            GREEN_CONFIG=str(self.env_filename), HOME=str(self.tmpd)
        ):
            cfg = config.getConfig()
            self.assertEqual(["green"], cfg.sections())
            self.assertEqual(
                str(self.default_filename), cfg.get("green", "omit-patterns")
            )
            self.assertRaises(
                configparser.NoOptionError, cfg.get, "green", "run-coverage"
            )
            self.assertEqual(self.default_logging, cfg.getboolean("green", "logging"))
            self.assertEqual(
                self.env_no_skip_report, cfg.getboolean("green", "no-skip-report")
            )
            self.assertEqual(self.default_version, cfg.getboolean("green", "version"))
            self.assertRaises(
                configparser.NoOptionError, cfg.getint, "green", "verbose"
            )

    def test_nocmd_env_def_nosetup(self):
        """
        Setup: no --config option, $GREEN_CONFIG is set, $HOME/.green exists,
            setup.cfg does not exist
        Result: load $GREEN_CONFIG
        """
        self.cmd_filename.unlink(missing_ok=True)
        self.setup_filename.unlink(missing_ok=True)
        with ModifiedEnvironment(
            GREEN_CONFIG=str(self.env_filename), HOME=str(self.tmpd)
        ):
            cfg = config.getConfig()
            self.assertEqual(["green"], cfg.sections())
            self.assertEqual(str(self.env_filename), cfg.get("green", "omit-patterns"))
            self.assertRaises(
                configparser.NoOptionError, cfg.get, "green", "run-coverage"
            )
            self.assertEqual(self.env_logging, cfg.getboolean("green", "logging"))
            self.assertEqual(
                self.env_no_skip_report, cfg.getboolean("green", "no-skip-report")
            )
            self.assertEqual(self.default_version, cfg.getboolean("green", "version"))
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "verbose"
            )

    def test_nocmd_env_nodef_nosetup(self):
        """
        Setup: no --config option, $GREEN_CONFIG is set, $HOME/.green does not
            exist, setup.cfg does not exist
        Result: load $GREEN_CONFIG
        """
        self.cmd_filename.unlink(missing_ok=True)
        self.default_filename.unlink(missing_ok=True)
        self.setup_filename.unlink(missing_ok=True)
        with ModifiedEnvironment(
            GREEN_CONFIG=str(self.env_filename), HOME=str(self.tmpd)
        ):
            cfg = config.getConfig()
            self.assertEqual(["green"], cfg.sections())
            self.assertEqual(str(self.env_filename), cfg.get("green", "omit-patterns"))
            self.assertRaises(
                configparser.NoOptionError, cfg.get, "green", "run-coverage"
            )
            self.assertEqual(self.env_logging, cfg.getboolean("green", "logging"))
            self.assertEqual(
                self.env_no_skip_report, cfg.getboolean("green", "no-skip-report")
            )
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "version"
            )
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "verbose"
            )

    def test_nocmd_noenv_def_nosetup(self):
        """
        Setup: no --config option, $GREEN_CONFIG unset, $HOME/.green exists,
            setup.cfg does not exist
        Result: load $HOME/.green
        """
        os.unlink(self.cmd_filename)
        os.unlink(self.env_filename)
        os.remove(self.setup_filename)
        with ModifiedEnvironment(GREEN_CONFIG=None, HOME=str(self.tmpd)):
            cfg = config.getConfig()
            self.assertEqual(["green"], cfg.sections())
            self.assertEqual(
                str(self.default_filename), cfg.get("green", "omit-patterns")
            )
            self.assertRaises(
                configparser.NoOptionError, cfg.get, "green", "run-coverage"
            )
            self.assertEqual(self.default_logging, cfg.getboolean("green", "logging"))
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "no-skip-report"
            )
            self.assertEqual(self.default_version, cfg.getboolean("green", "version"))
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "verbose"
            )

    def test_nocmd_noenv_nodef_nosetup(self):
        """
        Setup: no --config option, $GREEN_CONFIG unset, no $HOME/.green,
            setup.cfg does not exist
        Result: empty config
        """
        os.unlink(self.default_filename)
        os.unlink(self.env_filename)
        os.unlink(self.cmd_filename)
        os.remove(self.setup_filename)
        with ModifiedEnvironment(GREEN_CONFIG=None, HOME=str(self.tmpd)):
            cfg = config.getConfig()
            self.assertEqual([], cfg.sections())
            self.assertRaises(
                configparser.NoSectionError, cfg.get, "green", "omit-patterns"
            )
            self.assertRaises(
                configparser.NoSectionError, cfg.get, "green", "run-coverage"
            )
            self.assertRaises(configparser.NoSectionError, cfg.get, "green", "logging")
            self.assertRaises(
                configparser.NoSectionError, cfg.get, "green", "no-skip-report"
            )
            self.assertRaises(configparser.NoSectionError, cfg.get, "green", "version")
            self.assertRaises(configparser.NoSectionError, cfg.get, "green", "verbose")

    def test_cmd_env_nodef_setup(self):
        """
        Setup: --config on cmd, $GREEN_CONFIG is set, $HOME/.green does not
            exist, setup.cfg exists
        Result: load --config
        """
        os.unlink(self.default_filename)
        with ModifiedEnvironment(
            GREEN_CONFIG=str(self.env_filename), HOME=str(self.tmpd)
        ):
            cfg = config.getConfig(self.cmd_filename)
            self.assertEqual(["green"], cfg.sections())
            self.assertEqual(str(self.cmd_filename), cfg.get("green", "omit-patterns"))
            self.assertEqual(
                self.cmd_run_coverage, cfg.getboolean("green", "run-coverage")
            )
            self.assertEqual(self.cmd_logging, cfg.getboolean("green", "logging"))
            self.assertEqual(
                self.env_no_skip_report, cfg.getboolean("green", "no-skip-report")
            )
            self.assertEqual(self.setup_verbose, cfg.getint("green", "verbose"))
            self.assertEqual(self.setup_failfast, cfg.getboolean("green", "failfast"))
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "version"
            )

    def test_cmd_noenv_def_setup(self):
        """
        Setup: --config on cmd, $GREEN_CONFIG unset, $HOME/.green exists,
            setup.cfg exists
        Result: load --config
        """
        os.unlink(self.env_filename)
        with ModifiedEnvironment(GREEN_CONFIG=None, HOME=str(self.tmpd)):
            cfg = config.getConfig(self.cmd_filename)
            self.assertEqual(["green"], cfg.sections())
            self.assertEqual(str(self.cmd_filename), cfg.get("green", "omit-patterns"))
            self.assertEqual(
                self.cmd_run_coverage, cfg.getboolean("green", "run-coverage")
            )
            self.assertEqual(self.cmd_logging, cfg.getboolean("green", "logging"))
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "no-skip-report"
            )
            self.assertEqual(self.default_version, cfg.getboolean("green", "version"))
            self.assertEqual(self.setup_verbose, cfg.getint("green", "verbose"))
            self.assertEqual(self.setup_failfast, cfg.getboolean("green", "failfast"))

    def test_cmd_noenv_nodef_setup(self):
        """
        Setup: --config on cmd, $GREEN_CONFIG unset, $HOME/.green does not exist,
            setup.cfg exists
        Result: load --config
        """
        os.unlink(self.env_filename)
        os.unlink(self.default_filename)
        with ModifiedEnvironment(GREEN_CONFIG=None, HOME=str(self.tmpd)):
            cfg = config.getConfig(self.cmd_filename)
            self.assertEqual(["green"], cfg.sections())
            self.assertEqual(str(self.cmd_filename), cfg.get("green", "omit-patterns"))
            self.assertEqual(
                self.cmd_run_coverage, cfg.getboolean("green", "run-coverage")
            )
            self.assertEqual(self.cmd_logging, cfg.getboolean("green", "logging"))
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "no-skip-report"
            )
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "version"
            )
            self.assertEqual(self.setup_verbose, cfg.getint("green", "verbose"))
            self.assertEqual(self.setup_failfast, cfg.getboolean("green", "failfast"))

    def test_nocmd_env_def_setup(self):
        """
        Setup: no --config option, $GREEN_CONFIG is set, $HOME/.green exists,
            setup.cfg exists
        Result: load $GREEN_CONFIG
        """
        os.unlink(self.cmd_filename)
        with ModifiedEnvironment(
            GREEN_CONFIG=str(self.env_filename), HOME=str(self.tmpd)
        ):
            cfg = config.getConfig()
            self.assertEqual(["green"], cfg.sections())
            self.assertEqual(str(self.env_filename), cfg.get("green", "omit-patterns"))
            self.assertRaises(
                configparser.NoOptionError, cfg.get, "green", "run-coverage"
            )
            self.assertEqual(self.env_logging, cfg.getboolean("green", "logging"))
            self.assertEqual(
                self.env_no_skip_report, cfg.getboolean("green", "no-skip-report")
            )
            self.assertEqual(self.default_version, cfg.getboolean("green", "version"))
            self.assertEqual(self.setup_verbose, cfg.getint("green", "verbose"))
            self.assertEqual(self.setup_failfast, cfg.getboolean("green", "failfast"))

    def test_nocmd_env_nodef_setup(self):
        """
        Setup: no --config option, $GREEN_CONFIG is set, $HOME/.green does not
            exist, setup.cfg exists
        Result: load $GREEN_CONFIG
        """
        os.unlink(self.cmd_filename)
        os.unlink(self.default_filename)
        with ModifiedEnvironment(
            GREEN_CONFIG=str(self.env_filename), HOME=str(self.tmpd)
        ):
            cfg = config.getConfig()
            self.assertEqual(["green"], cfg.sections())
            self.assertEqual(str(self.env_filename), cfg.get("green", "omit-patterns"))
            self.assertRaises(
                configparser.NoOptionError, cfg.get, "green", "run-coverage"
            )
            self.assertEqual(self.env_logging, cfg.getboolean("green", "logging"))
            self.assertEqual(
                self.env_no_skip_report, cfg.getboolean("green", "no-skip-report")
            )
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "version"
            )
            self.assertEqual(self.setup_verbose, cfg.getint("green", "verbose"))
            self.assertEqual(self.setup_failfast, cfg.getboolean("green", "failfast"))

    def test_nocmd_noenv_def_setup(self):
        """
        Setup: no --config option, $GREEN_CONFIG unset, $HOME/.green exists,
            setup.cfg exists
        Result: load $HOME/.green
        """
        os.unlink(self.cmd_filename)
        os.unlink(self.env_filename)
        with ModifiedEnvironment(GREEN_CONFIG=None, HOME=str(self.tmpd)):
            cfg = config.getConfig()
            self.assertEqual(["green"], cfg.sections())
            self.assertEqual(
                str(self.default_filename), cfg.get("green", "omit-patterns")
            )
            self.assertRaises(
                configparser.NoOptionError, cfg.get, "green", "run-coverage"
            )
            self.assertEqual(self.default_logging, cfg.getboolean("green", "logging"))
            self.assertRaises(
                configparser.NoOptionError, cfg.getboolean, "green", "no-skip-report"
            )
            self.assertEqual(self.default_version, cfg.getboolean("green", "version"))
            self.assertEqual(self.setup_verbose, cfg.getint("green", "verbose"))
            self.assertEqual(self.setup_failfast, cfg.getboolean("green", "failfast"))

    def test_nocmd_noenv_nodef_setup(self):
        """
        Setup: no --config option, $GREEN_CONFIG unset, no $HOME/.green,
            setup.cfg exists
        Result: empty config
        """
        self.default_filename.unlink(missing_ok=True)
        self.env_filename.unlink(missing_ok=True)
        self.cmd_filename.unlink(missing_ok=True)
        with ModifiedEnvironment(GREEN_CONFIG=None, HOME=str(self.tmpd)):
            cfg = config.getConfig()
            self.assertEqual(self.setup_verbose, cfg.getint("green", "verbose"))
            self.assertEqual(self.setup_failfast, cfg.getboolean("green", "failfast"))
            self.assertRaises(
                configparser.NoOptionError, cfg.get, "green", "omit-patterns"
            )
            self.assertRaises(
                configparser.NoOptionError, cfg.get, "green", "run-coverage"
            )
            self.assertRaises(configparser.NoOptionError, cfg.get, "green", "logging")
            self.assertRaises(
                configparser.NoOptionError, cfg.get, "green", "no-skip-report"
            )
            self.assertRaises(configparser.NoOptionError, cfg.get, "green", "version")


class TestMergeConfig(ConfigBase):
    """
    Merging config files and command-line arguments works as expected.
    """

    def test_overwrite(self):
        """
        Non-default command-line argument values overwrite config values.
        """
        # This config environment should set the values we look at to False and
        # a filename in omit-patterns
        s = StringIO()
        gs = GreenStream(s)
        saved_stdout = config.sys.stdout
        config.sys.stdout = gs
        self.addCleanup(setattr, config.sys, "stdout", saved_stdout)
        with ModifiedEnvironment(
            GREEN_CONFIG=str(self.env_filename), HOME=str(self.tmpd)
        ):
            new_args = copy.deepcopy(config.get_default_args())

            new_args.omit_patterns = "omitstuff"
            new_args.run_coverage = True
            new_args.logging = True
            new_args.no_skip_report = True
            new_args.version = True

            new_args.config = self.cmd_filename
            computed_args = config.mergeConfig(new_args, testing=True)

            self.assertEqual(computed_args.omit_patterns, "omitstuff")
            self.assertEqual(computed_args.run_coverage, new_args.run_coverage)
            self.assertEqual(computed_args.logging, new_args.logging)
            self.assertEqual(computed_args.no_skip_report, new_args.no_skip_report)
            self.assertEqual(computed_args.version, new_args.version)

    def test_no_overwrite(self):
        """
        Default unspecified command-line args do not overwrite config values.
        """
        # This config environment should set logging to True
        with ModifiedEnvironment(GREEN_CONFIG=str(self.env_filename), HOME=""):
            # The default for logging in arguments is False
            da = copy.deepcopy(config.get_default_args())
            del da.logging
            computed_args = config.mergeConfig(da, testing=True)
            self.assertEqual(computed_args.logging, True)

    def test_specified_command_line(self):
        """
        Specified command-line arguments always overwrite config file values
        """
        with ModifiedEnvironment(HOME=str(self.tmpd)):
            new_args = copy.deepcopy(config.get_default_args())
            new_args.failfast = True  # same as config, for sanity
            new_args.logging = True  # different than config, not default
            del new_args.version  # Not in arguments, should get config value
            new_args.termcolor = False  # override config, set back to default
            computed_args = config.mergeConfig(new_args, testing=True)
            self.assertEqual(computed_args.failfast, True)
            self.assertEqual(computed_args.logging, True)
            self.assertEqual(computed_args.version, False)
            self.assertEqual(computed_args.termcolor, False)

    def test_targets(self):
        """
        The targets passed in make it through mergeConfig, and the specified
        target gets parsed
        """
        config.sys.argv = ["", "target1", "target2"]
        args = config.parseArguments()
        args = config.mergeConfig(args)
        self.assertEqual(args.targets, ["target1", "target2"])

    def test_forgot_to_update_merge(self):
        """
        mergeConfig raises an exception for unknown cmdline args
        """
        default_args = config.get_default_args()
        orig_args = copy.deepcopy(default_args)
        self.addCleanup(setattr, config, "default_args", orig_args)

        default_args.new_option = True
        new_args = copy.deepcopy(default_args)

        try:
            with self.assertRaises(NotImplementedError):
                config.mergeConfig(new_args, testing=True)
        finally:
            config.get_default_args.cache_clear()



"""Classes and methods to generate JUnit XML reports."""

from __future__ import annotations

from typing import Dict, List, Final, TextIO, Tuple, TYPE_CHECKING, Union

from lxml.etree import Element, tostring as to_xml

if TYPE_CHECKING:
    # TypeAlias moved to the typing module after py3.9.
    from typing_extensions import TypeAlias

    from green.result import GreenTestResult, ProtoTest, ProtoError
    from lxml.etree import _Element

    # TODO: use NamedTuple for TestVerdict.
    TestVerdict: TypeAlias = Union[
        Tuple[int, ProtoTest], Tuple[int, ProtoTest, Union[str, ProtoError]]
    ]
    TestsCollection: TypeAlias = Dict[str, List[TestVerdict]]


# TODO: consider using enum.Enum (new in py3.4) for the JUnitDialect and Verdict classes.


class JUnitDialect:
    """
    Hold the name of the elements defined in the JUnit XML schema (for JUnit 4).
    """

    CLASS_NAME: Final[str] = "classname"
    ERROR: Final[str] = "error"
    ERROR_COUNT: Final[str] = "errors"
    FAILURE: Final[str] = "failure"
    FAILURE_COUNT: Final[str] = "failures"
    NAME: Final[str] = "name"
    SKIPPED: Final[str] = "skipped"
    SKIPPED_COUNT: Final[str] = "skipped"
    SYSTEM_ERR: Final[str] = "system-err"
    SYSTEM_OUT: Final[str] = "system-out"
    TEST_CASE: Final[str] = "testcase"
    TEST_COUNT: Final[str] = "tests"
    TEST_SUITE: Final[str] = "testsuite"
    TEST_SUITES: Final[str] = "testsuites"
    TEST_TIME: Final[str] = "time"


class Verdict:
    """
    Enumeration of possible test verdicts
    """

    PASSED: Final[int] = 0
    FAILED: Final[int] = 1
    ERROR: Final[int] = 2
    SKIPPED: Final[int] = 3


class JUnitXML:
    """
    Serialize a GreenTestResult object into a JUnit XML file, that can
    be read by continuous integration servers, for example.

    See GitHub Issue #104
    See Option '-j' / '--junit-report'
    """

    def save_as(self, test_results: GreenTestResult, destination: TextIO) -> None:
        """
        Write the JUnit XML report to the given file-like object.
        """
        xml_root = Element(JUnitDialect.TEST_SUITES)
        tests_by_class = self._group_tests_by_class(test_results)
        suite: list[TestVerdict]
        for name, suite in tests_by_class.items():
            xml_suite = self._convert_suite(test_results, name, suite)
            xml_root.append(xml_suite)

        xml_root.set(JUnitDialect.TEST_TIME, str(test_results.timeTaken))

        xml = to_xml(
            xml_root,
            xml_declaration=True,
            pretty_print=True,
            encoding="utf-8",
            method="xml",
        )
        destination.write(xml.decode())

    def _group_tests_by_class(
        self, test_results: GreenTestResult
    ) -> dict[str, list[TestVerdict]]:
        result: TestsCollection = {}
        self._add_passing_tests(result, test_results)
        self._add_failures(result, test_results)
        self._add_errors(result, test_results)
        self._add_skipped_tests(result, test_results)
        return result

    @staticmethod
    def _add_passing_tests(
        collection: TestsCollection, test_results: GreenTestResult
    ) -> None:
        for each_test in test_results.passing:
            key = JUnitXML._suite_name(each_test)
            if key not in collection:
                collection[key] = []
            collection[key].append((Verdict.PASSED, each_test))

    @staticmethod
    def _suite_name(test) -> str:
        return f"{test.module}.{test.class_name}"

    @staticmethod
    def _add_failures(
        collection: TestsCollection, test_results: GreenTestResult
    ) -> None:
        for each_test, failure in test_results.failures:
            key = JUnitXML._suite_name(each_test)
            if key not in collection:
                collection[key] = []
            collection[key].append((Verdict.FAILED, each_test, failure))

    @staticmethod
    def _add_errors(collection: TestsCollection, test_results: GreenTestResult):
        for each_test, error in test_results.errors:
            key = JUnitXML._suite_name(each_test)
            if key not in collection:
                collection[key] = []
            collection[key].append((Verdict.ERROR, each_test, error))

    @staticmethod
    def _add_skipped_tests(
        collection: TestsCollection, test_results: GreenTestResult
    ) -> None:
        for each_test, reason in test_results.skipped:
            key = JUnitXML._suite_name(each_test)
            if key not in collection:
                collection[key] = []
            collection[key].append((Verdict.SKIPPED, each_test, reason))

    def _convert_suite(
        self, results: GreenTestResult, name: str, suite: list[TestVerdict]
    ) -> _Element:
        xml_suite = Element(JUnitDialect.TEST_SUITE)
        xml_suite.set(JUnitDialect.NAME, name)
        xml_suite.set(JUnitDialect.TEST_COUNT, str(len(suite)))
        xml_suite.set(
            JUnitDialect.FAILURE_COUNT,
            str(self._count_test_with_verdict(Verdict.FAILED, suite)),
        )
        xml_suite.set(
            JUnitDialect.ERROR_COUNT,
            str(self._count_test_with_verdict(Verdict.ERROR, suite)),
        )
        xml_suite.set(
            JUnitDialect.SKIPPED_COUNT,
            str(self._count_test_with_verdict(Verdict.SKIPPED, suite)),
        )
        xml_suite.set(JUnitDialect.TEST_TIME, str(self._suite_time(suite)))
        for each_test in suite:
            xml_test = self._convert_test(results, *each_test)
            xml_suite.append(xml_test)

        return xml_suite

    @staticmethod
    def _count_test_with_verdict(verdict: int, suite: list[TestVerdict]) -> int:
        return sum(1 for entry in suite if entry[0] == verdict)

    def _convert_test(
        self,
        results: GreenTestResult,
        verdict: int,
        test: ProtoTest,
        *details: str | ProtoError,
    ) -> _Element:
        xml_test = Element(JUnitDialect.TEST_CASE)
        xml_test.set(JUnitDialect.NAME, test.method_name)
        xml_test.set(JUnitDialect.CLASS_NAME, test.class_name)
        xml_test.set(JUnitDialect.TEST_TIME, test.test_time)

        error: str | ProtoError | None = details[0] if details else None
        xml_verdict = self._convert_verdict(verdict, test, error)
        if xml_verdict is not None:
            xml_test.append(xml_verdict)

        if test in results.stdout_output:
            system_out = Element(JUnitDialect.SYSTEM_OUT)
            system_out.text = results.stdout_output[test]
            xml_test.append(system_out)

        if test in results.stderr_errput:
            system_err = Element(JUnitDialect.SYSTEM_ERR)
            system_err.text = results.stderr_errput[test]
            xml_test.append(system_err)

        return xml_test

    # FIXME: test is not used.
    def _convert_verdict(
        self, verdict: int, test: ProtoTest, error_details: str | ProtoError | None
    ) -> _Element | None:
        message = str(error_details) if error_details else ""
        if verdict == Verdict.FAILED:
            failure = Element(JUnitDialect.FAILURE)
            failure.text = message
            return failure
        if verdict == Verdict.ERROR:
            error = Element(JUnitDialect.ERROR)
            error.text = message
            return error
        if verdict == Verdict.SKIPPED:
            skipped = Element(JUnitDialect.SKIPPED)
            skipped.text = message
            return skipped
        return None

    @staticmethod
    def _suite_time(suite: list[TestVerdict]) -> float:
        return sum(float(each_test.test_time) for verdict, each_test, *details in suite)



"""Classes and methods to handle test results."""

from __future__ import annotations

import argparse
from doctest import DocTest, DocTestCase
from math import ceil
from shutil import get_terminal_size
import time
import traceback
from typing import Any, Callable, Sequence, TYPE_CHECKING, Union
from unittest.result import failfast
from unittest import TestCase, TestSuite

from green.output import Colors, debug, GreenStream
from green.version import pretty_version

if TYPE_CHECKING:
    from green.process import ExcInfoType

    TestCaseT = Union["ProtoTest", TestCase, DocTestCase]
    RunnableTestT = Union[TestCaseT, TestSuite]

terminal_width, _ignored = get_terminal_size()


def proto_test(test: RunnableTestT) -> ProtoTest:
    """
    If test is a ProtoTest, I just return it.
    Otherwise, I create a ProtoTest out of test and return it.
    """
    if isinstance(test, ProtoTest):
        return test
    return ProtoTest(test)


def proto_error(err: ExcInfoType | ProtoError) -> ProtoError:
    """
    If err is a ProtoError, I just return it.
    Otherwise, I create a ProtoError out of err and return it.
    """
    if isinstance(err, ProtoError):
        return err
    return ProtoError(err)


class ProtoTest:
    """
    I take a full-fledged TestCase and preserve just the information we need
    and can pass between processes.
    """

    module: str = ""
    class_name: str = ""
    method_name: str = ""
    docstr_part: str = ""
    subtest_part: str = ""
    test_time: str = "0.0"
    failureException = AssertionError
    description: str = ""

    # Doctests specific attributes:
    is_doctest: bool = False
    filename: str | None = None
    name: str = ""

    def __init__(self, test: TestCase | DocTestCase | TestSuite | None = None) -> None:
        # Teardown handling is a royal mess
        self.is_class_or_module_teardown_error: bool = False

        # Is this a subtest? The _SubTest class is private so we need to check the attributes.
        sub_description = getattr(test, "_subDescription", None)
        if sub_description is not None:
            self.subtest_part = " " + sub_description()
            test = test.test_case  # type: ignore

        # Is this a DocTest?
        # We need to know that this is a doctest, because doctests are very
        # different than regular test cases in many ways, so they get special
        # treatment inside and outside of this class.
        if isinstance(test, DocTestCase):
            self.is_doctest = True
            dt_test: DocTest = test._dt_test  # type: ignore
            self.name = dt_test.name
            # We had to inject this in green/loader.py -- this is the test
            # module that specified that we should load doctests from some
            # other module -- so that we'll group the doctests with the test
            # module that specified that we should load them.
            self.module = test.__module__
            self.class_name = "DocTests via `doctest_modules = [...]`"
            # I'm not sure this will be the correct way to get the method name
            # in all cases.
            self.method_name = self.name.split(".")[1]
            self.filename = dt_test.filename
            self.lineno = dt_test.lineno

        # Is this a TestCase?
        elif isinstance(test, TestCase):
            self.module = test.__module__
            self.class_name = test.__class__.__name__
            self.method_name = str(test).split()[0]
            # docstr_part strips initial whitespace, then combines all lines
            # into one string until the first completely blank line in the
            # docstring
            doc_segments = []
            if getattr(test, "_testMethodDoc", None):
                for line in test._testMethodDoc.lstrip().split("\n"):
                    line = line.strip()
                    if not line:
                        break
                    doc_segments.append(line)
            self.docstr_part = " ".join(doc_segments)

    def __eq__(self, other: Any) -> bool:
        return self.__hash__() == other.__hash__()

    def __hash__(self) -> int:
        return hash(self.dotted_name)

    def __str__(self) -> str:
        return self.dotted_name

    @property
    def dotted_name(self, ignored: Any = None) -> str:
        if self.is_doctest or self.is_class_or_module_teardown_error:
            return self.name
        return f"{self.module}.{self.class_name}.{self.method_name}{self.subtest_part}"

    def getDescription(self, verbose: int) -> str:
        # Classes or module teardown errors
        if self.is_class_or_module_teardown_error:
            return self.name
        # Doctests
        if self.is_doctest:
            if verbose == 2:
                return self.name
            if verbose > 2:
                return f"{self.name} -> {self.filename}:{self.lineno}"
            return ""
        # Regular tests
        if verbose == 2:
            return f"{self.method_name}{self.subtest_part}"
        elif verbose == 3:
            return f"{self.docstr_part}{self.subtest_part}" or self.method_name
        elif verbose > 3:
            if self.docstr_part + self.subtest_part:
                return f"{self.method_name}: {self.docstr_part}{self.subtest_part}"
            return self.method_name
        return ""


class ProtoError:
    """
    I take a full-fledged test error and preserve just the information we need
    and can pass between processes.
    """

    def __init__(self, err: ExcInfoType) -> None:
        self.traceback_lines = traceback.format_exception(*err)

    def __str__(self) -> str:
        return "\n".join(self.traceback_lines)


class BaseTestResult:
    """
    I am inherited by ProtoTestResult and GreenTestResult.
    """

    def __init__(self, stream: GreenStream | None, *, colors: Colors | None = None):
        self.stdout_output: dict[ProtoTest, str] = {}
        self.stderr_errput: dict[ProtoTest, str] = {}
        self.stream: GreenStream | None = stream
        self.colors: Colors = colors or Colors()
        # The collectedDurations list is new in Python 3.12.
        self.collectedDurations: list[tuple[str, float]] = []

    def recordStdout(self, test: RunnableTestT, output):
        """
        Called with stdout that the suite decided to capture so we can report
        the captured output somewhere.
        """
        if output:
            test = proto_test(test)
            self.stdout_output[test] = output

    def recordStderr(self, test: RunnableTestT, errput):
        """
        Called with stderr that the suite decided to capture so we can report
        the captured "errput" somewhere.
        """
        if errput:
            test = proto_test(test)
            self.stderr_errput[test] = errput

    def displayStdout(self, test: TestCaseT):
        """
        Displays AND REMOVES the output captured from a specific test.  The
        removal is done so that this method can be called multiple times
        without duplicating results output.
        """
        test = proto_test(test)
        if test.dotted_name in self.stdout_output:
            if self.stream is None:
                raise ValueError("stream is None")
            colors = self.colors
            captured = "Captured stdout"
            self.stream.write(
                f"\n{colors.yellow(captured)} for {colors.bold(test.dotted_name)}\n"
                f"{self.stdout_output[test]}"
            )
            del self.stdout_output[test]

    def displayStderr(self, test: TestCaseT):
        """
        Displays AND REMOVES the errput captured from a specific test.  The
        removal is done so that this method can be called multiple times
        without duplicating results errput.
        """
        test = proto_test(test)
        if test.dotted_name in self.stderr_errput:
            if self.stream is None:
                raise ValueError("stream is None")
            colors = self.colors
            captured = "Captured stderr"
            self.stream.write(
                f"\n{colors.yellow(captured)} for {colors.bold(test.dotted_name)}\n"
                f"{self.stderr_errput[test]}"
            )
            del self.stderr_errput[test]

    def addDuration(self, test: TestCaseT, elapsed: float):
        """
        Called when a test finished running, regardless of its outcome.

        New in Python 3.12.

        Args:
            test: The test case corresponding to the test method.
            elapsed: The time represented in seconds, including the
                execution of cleanup functions.
        """
        self.collectedDurations.append((str(test), elapsed))


class ProtoTestResult(BaseTestResult):
    """
    I'm the TestResult object for a single unit test run in a process.
    """

    failfast: bool = False  # Because unittest inspects the attribute
    finalize_callback_called: bool = False
    shouldStop: bool = False
    start_time: float = 0.0
    test_time: str = ""

    def __init__(
        self,
        start_callback: Callable[[RunnableTestT], None] | None = None,
        finalize_callback: Callable[[ProtoTestResult], None] | None = None,
    ) -> None:
        super().__init__(None, colors=None)
        self.start_callback = start_callback
        self.finalize_callback = finalize_callback
        self.collectedDurations: list[tuple[str, float]] = []
        self.errors: list[tuple[ProtoTest, ProtoError]] = []
        self.expectedFailures: list[tuple[ProtoTest, ProtoError]] = []
        self.failures: list[tuple[ProtoTest, ProtoError]] = []
        self.passing: list[ProtoTest] = []
        self.skipped: list[tuple[ProtoTest, str]] = []
        self.unexpectedSuccesses: list[ProtoTest] = []
        self.pickle_attrs: Sequence[str] = (
            "errors",
            "expectedFailures",
            "failures",
            "passing",
            "pickle_attrs",  # TODO: check if pickle_attrs should be pickled.
            "shouldStop",
            "skipped",
            "stderr_errput",
            "stdout_output",
            "test_time",
            "unexpectedSuccesses",
        )
        self.reinitialize()

    def reinitialize(self):
        self.shouldStop = False
        self.collectedDurations.clear()
        self.errors.clear()
        self.expectedFailures.clear()
        self.failures.clear()
        self.passing.clear()
        self.skipped.clear()
        self.unexpectedSuccesses.clear()
        self.start_time = 0.0
        self.test_time = ""

    def __repr__(self) -> str:  # pragma: no cover
        return (
            f"errors{self.errors},"
            f" expectedFailures{self.expectedFailures},"
            f" failures{self.failures},"
            f" passing{self.passing},"
            f" skipped{self.skipped},"
            f" unexpectedSuccesses{self.unexpectedSuccesses},"
            f" test_time{self.test_time}"
        )

    def __getstate__(self) -> dict[str, Any]:
        """
        Prevent the callback functions from getting pickled.
        """
        result_dict = {}
        for pickle_attr in self.pickle_attrs:
            result_dict[pickle_attr] = self.__dict__[pickle_attr]
        return result_dict

    def __setstate__(self, state: dict[str, Any]) -> None:
        """
        Since the callback functions weren't pickled, we need to initialize them.
        """
        self.__dict__.update(state)
        self.start_callback = None
        self.finalize_callback = None

    def startTest(self, test: RunnableTestT) -> None:
        """
        Called before each test runs.
        """
        test = proto_test(test)
        self.start_time = time.time()
        self.reinitialize()
        if self.start_callback:
            self.start_callback(test)

    def stopTest(self, test: RunnableTestT) -> None:
        """
        Called after each test runs.
        """
        if self.start_time:
            self.test_time = str(time.time() - self.start_time)
        else:
            self.test_time = "0.0"

    def finalize(self) -> None:
        """
        I am here so that after the GreenTestSuite has had a chance to inject
        the captured stdout/stderr back into me, I can relay that through to
        the worker process's poolRunner who will send me back up to the parent
        process.
        """
        if self.finalize_callback:
            self.finalize_callback(self)
            self.finalize_callback_called = True

    def addSuccess(self, test: TestCaseT) -> None:
        """
        Called when a test passed.
        """
        self.passing.append(proto_test(test))

    def addError(self, test: RunnableTestT, err: ProtoError | ExcInfoType) -> None:
        """
        Called when a test raises an exception.
        """
        self.errors.append((proto_test(test), proto_error(err)))

    def addFailure(self, test: TestCaseT, err: ExcInfoType) -> None:
        """
        Called when a test fails a unittest assertion.
        """
        self.failures.append((proto_test(test), proto_error(err)))

    def addSkip(self, test: TestCaseT, reason: str) -> None:
        """
        Called when a test is skipped.
        """
        self.skipped.append((proto_test(test), reason))

    def addExpectedFailure(self, test: TestCaseT, err: ExcInfoType) -> None:
        """
        Called when a test fails, and we expected the failure.
        """
        self.expectedFailures.append((proto_test(test), proto_error(err)))

    def addUnexpectedSuccess(self, test: TestCaseT) -> None:
        """
        Called when a test passed, but we expected a failure
        """
        self.unexpectedSuccesses.append(proto_test(test))

    # The _SubTest class is private and masked so we cannot easily type annotate.
    def addSubTest(
        self, test: TestCaseT, subtest: Any, err: ExcInfoType | None
    ) -> None:
        """
        Called at the end of a subtest no matter its result.

        The test that runs the subtests calls the other test methods to
        record its own result.  We use this method to record each subtest as a
        separate test result.  It's very meta.
        """
        if err is not None:
            if err[0] is not None and issubclass(err[0], test.failureException):
                self.addFailure(subtest, err)
            else:
                self.addError(subtest, err)


class GreenTestResult(BaseTestResult):
    """
    Aggregates test results and outputs them to a stream.
    """

    stream: GreenStream
    last_class: str = ""
    last_module: str = ""
    first_text_output: str = ""
    shouldStop: bool = False

    def __init__(self, args: argparse.Namespace, stream: GreenStream) -> None:
        super().__init__(stream, colors=Colors(args.termcolor))
        self.args = args
        self.showAll: bool = args.verbose > 1
        self.dots: bool = args.verbose == 1
        self.verbose: int = args.verbose
        self.failfast = args.failfast
        self.testsRun: int = 0
        # Individual lists
        self.collectedDurations: list[tuple[str, float]] = []
        self.errors: list[tuple[ProtoTest, ProtoError]] = []
        self.expectedFailures: list[tuple[ProtoTest, ProtoError]] = []
        self.failures: list[tuple[ProtoTest, ProtoError]] = []
        self.passing: list[ProtoTest] = []
        self.skipped: list[tuple[ProtoTest, str]] = []
        self.unexpectedSuccesses: list[ProtoTest] = []
        # Combination of all errors and failures
        self.all_errors: list[
            tuple[ProtoTest, Callable[[str], str], str, ProtoError]
        ] = []
        # For exiting non-zero if we don't reach a certain level of coverage
        self.coverage_percent: int | None = None

    def __str__(self) -> str:  # pragma: no cover
        return (
            f"tests run: {self.testsRun}, "
            f"errors{self.errors}, "
            f"expectedFailures{self.expectedFailures}, "
            f"failures{self.failures}, "
            f"passing{self.passing}, "
            f"skipped{self.skipped}, "
            f"unexpectedSuccesses{self.unexpectedSuccesses}"
        )

    def stop(self) -> None:
        self.shouldStop = True

    def tryRecordingStdoutStderr(
        self,
        test: ProtoTest,
        proto_test_result: ProtoTestResult,
        err: ProtoError | None = None,
    ) -> None:
        if proto_test_result.stdout_output.get(test, False):
            self.recordStdout(test, proto_test_result.stdout_output[test])
        if proto_test_result.stderr_errput.get(test, False):
            self.recordStderr(test, proto_test_result.stderr_errput[test])

        # SubTest errors/failures (but not successes) generate a different err object, so we have to
        # do some inspection to figure out which object has the output/errput
        if (test.class_name == "SubTest") and err:
            for t in proto_test_result.stdout_output.keys():
                if test.dotted_name.startswith(t.dotted_name):
                    self.recordStdout(test, proto_test_result.stdout_output[t])
                    break
            for t in proto_test_result.stderr_errput.keys():
                if test.dotted_name.startswith(t.dotted_name):
                    self.recordStderr(test, proto_test_result.stderr_errput[t])
                    break

    def addProtoTestResult(self, proto_test_result: ProtoTestResult) -> None:
        for test, err in proto_test_result.errors:
            self.addError(test, err, proto_test_result.test_time)
            self.tryRecordingStdoutStderr(test, proto_test_result, err)
        for test, err in proto_test_result.expectedFailures:
            self.addExpectedFailure(test, err, proto_test_result.test_time)
            self.tryRecordingStdoutStderr(test, proto_test_result, err)
        for test, err in proto_test_result.failures:
            self.addFailure(test, err, proto_test_result.test_time)
            self.tryRecordingStdoutStderr(test, proto_test_result, err)
        for test in proto_test_result.passing:
            self.addSuccess(test, proto_test_result.test_time)
            self.tryRecordingStdoutStderr(test, proto_test_result)
        for test, reason in proto_test_result.skipped:
            self.addSkip(test, reason, proto_test_result.test_time)
            self.tryRecordingStdoutStderr(test, proto_test_result)
        for test in proto_test_result.unexpectedSuccesses:
            self.addUnexpectedSuccess(test, proto_test_result.test_time)
            self.tryRecordingStdoutStderr(test, proto_test_result)

    def startTestRun(self) -> None:
        """
        Called once before any tests run.
        """
        self.startTime = time.time()
        # Really verbose information
        if self.verbose > 2:
            self.stream.writeln(self.colors.bold(f"{pretty_version()}\n"))

    def stopTestRun(self) -> None:
        """
        Called once after all tests have run.
        """
        # FIXME: stopTime and timeTaken are defined outside __init__.
        self.stopTime = time.time()
        self.timeTaken = self.stopTime - self.startTime
        self.printErrors()
        if self.args.run_coverage or self.args.quiet_coverage:
            from coverage.misc import CoverageException

            try:
                self.stream.writeln()
                self.args.cov.stop()
                self.args.cov.save()
                self.args.cov.combine()
                self.args.cov.save()
                if not self.args.quiet_coverage:
                    self.stream.coverage_percent = None
                    self.args.cov.report(
                        file=self.stream,
                        omit=self.args.omit_patterns,
                        show_missing=True,
                        ignore_errors=True,
                    )
                    self.coverage_percent = self.stream.coverage_percent

            except CoverageException as ce:
                if (len(ce.args) == 1) and ("No data to report" not in ce.args[0]):
                    raise ce

        if self.testsRun and not self.shouldStop:
            self.stream.writeln()
        if self.shouldStop:
            self.stream.writeln()
            self.stream.writeln(
                self.colors.yellow("Warning: Some tests may not have been run.")
            )
            self.stream.writeln()
        self.stream.writeln(
            "Ran %s test%s in %ss using %s process%s"
            % (
                self.colors.bold(str(self.testsRun)),
                self.testsRun != 1 and "s" or "",
                self.colors.bold("%.3f" % self.timeTaken),
                self.colors.bold("%d" % self.args.processes),
                self.args.processes != 1 and "es" or "",
            )
        )
        self.stream.writeln()
        results: tuple[tuple[list, str, Callable[[str], str]], ...] = (
            (self.errors, "errors", self.colors.error),
            (self.expectedFailures, "expected_failures", self.colors.expectedFailure),
            (self.failures, "failures", self.colors.failing),
            (self.passing, "passes", self.colors.passing),
            (self.skipped, "skips", self.colors.skipped),
            (
                self.unexpectedSuccesses,
                "unexpected_successes",
                self.colors.unexpectedSuccess,
            ),
        )
        stats = []
        for obj_list, name, color_func in results:
            if obj_list:
                stats.append(f"{name}={color_func(str(len(obj_list)))}")
        if not stats:
            self.stream.writeln(self.colors.failing("No Tests Found"))
        else:
            grade = self.colors.passing("OK")
            if not self.wasSuccessful():
                grade = self.colors.failing("FAILED")
            self.stream.writeln(f"{grade} ({', '.join(stats)})")

    def startTest(self, test: RunnableTestT) -> None:
        """
        Called before the start of each test.
        """
        # Get our bearings
        test = proto_test(test)
        current_module = test.module
        current_class = test.class_name

        # Output
        if self.showAll:
            # Module...if it changed.
            if current_module != self.last_module:
                self.stream.writeln(self.colors.moduleName(current_module))
            # Class...if it changed.
            if current_class != self.last_class:
                self.stream.writeln(
                    self.colors.className(
                        self.stream.formatText(current_class, indent=1)
                    )
                )
            if self.stream.isatty():
                # In the terminal, we will write a placeholder, and then
                # modify the first character and rewrite it in color after
                # the test has run.
                self.first_text_output = self.stream.formatLine(
                    test.getDescription(self.verbose), indent=2
                )
                self.stream.write(self.colors.bold(self.first_text_output))
            self.stream.flush()

        # Set state for next time
        if current_module != self.last_module:
            self.last_module = current_module
        if current_class != self.last_class:
            self.last_class = current_class

    def stopTest(self, test: RunnableTestT) -> None:
        """
        Supposed to be called after each test.
        """

    def _reportOutcome(
        self,
        test: RunnableTestT,
        outcome_char: str,
        color_func: Callable[[str], str],
        err: ProtoError | None = None,
        reason: str = "",
    ) -> None:
        self.testsRun += 1
        test = proto_test(test)
        if self.showAll:
            if self.stream.isatty():
                self.stream.write(self.colors.start_of_line())
            # Can end up being different from the first time due to subtest
            # information only being available after a test result comes in.
            second_text_output = self.stream.formatLine(
                test.getDescription(self.verbose), indent=2, outcome_char=outcome_char
            )
            if self.stream.isatty() and terminal_width:  # pragma: no cover
                cursor_rewind = (
                    int(ceil(len(self.first_text_output) / terminal_width)) - 1
                )
                if cursor_rewind:
                    self.stream.write(self.colors.up(cursor_rewind))
            self.stream.write(color_func(second_text_output))
            if reason:
                self.stream.write(color_func(" -- " + reason))
            self.stream.writeln()
            self.stream.flush()
        elif self.dots:
            self.stream.write(color_func(outcome_char))
            self.stream.flush()

    def addSuccess(
        self, test: RunnableTestT, test_time: float | str | None = None
    ) -> None:
        """
        Called when a test passed.
        """
        test = proto_test(test)
        if test_time:
            test.test_time = str(test_time)
        self.passing.append(test)
        self._reportOutcome(test, ".", self.colors.passing)

    @failfast
    def addError(
        self, test: RunnableTestT, err: ProtoError, test_time: float | str | None = None
    ) -> None:
        """
        Called when a test raises an exception.
        """
        test = proto_test(test)
        if test_time:
            test.test_time = str(test_time)
        error = proto_error(err)
        self.errors.append((test, error))
        self.all_errors.append((test, self.colors.error, "Error", error))
        self._reportOutcome(test, "E", self.colors.error, error)

    @failfast
    def addFailure(
        self, test: RunnableTestT, err: ProtoError, test_time: float | str | None = None
    ) -> None:
        """
        Called when a test fails a unittest assertion.
        """
        # Special case: Catch Twisted's skips that come thtrough as failures
        # and treat them as skips instead
        if len(err.traceback_lines) == 1:
            if err.traceback_lines[0].startswith("UnsupportedTrialFeature"):
                reason = eval(err.traceback_lines[0][25:])[1]
                self.addSkip(test, reason)
                return

        test_proto = proto_test(test)
        if test_time:
            test_proto.test_time = str(test_time)
        self.failures.append((test_proto, err))
        self.all_errors.append((test_proto, self.colors.error, "Failure", err))
        self._reportOutcome(test_proto, "F", self.colors.failing, err)

    def addSkip(
        self, test: RunnableTestT, reason: str, test_time: float | str | None = None
    ) -> None:
        """
        Called when a test is skipped.
        """
        test = proto_test(test)
        if test_time:
            test.test_time = str(test_time)
        self.skipped.append((test, reason))
        self._reportOutcome(test, "s", self.colors.skipped, reason=reason)

    def addExpectedFailure(
        self, test: RunnableTestT, err: ProtoError, test_time: float | str | None = None
    ) -> None:
        """
        Called when a test fails, and we expected the failure.
        """
        test = proto_test(test)
        if test_time:
            test.test_time = str(test_time)
        err = proto_error(err)
        self.expectedFailures.append((test, err))
        self._reportOutcome(test, "x", self.colors.expectedFailure, err)

    def addUnexpectedSuccess(
        self, test: RunnableTestT, test_time: float | str | None = None
    ) -> None:
        """
        Called when a test passed, but we expected a failure.
        """
        test = proto_test(test)
        if test_time:
            test.test_time = str(test_time)
        self.unexpectedSuccesses.append(test)
        self._reportOutcome(test, "u", self.colors.unexpectedSuccess)

    def printErrors(self) -> None:
        """
        Print a list of all tracebacks from errors and failures, as well as
        captured stdout (even if the test passed, except with quiet_stdout
        option).
        """
        if self.dots:
            self.stream.writeln()

        # Skipped Test Report
        if not self.args.no_skip_report:
            for test, reason in self.skipped:
                self.stream.writeln(
                    "\n{} {} - {}".format(
                        self.colors.blue("Skipped"),
                        self.colors.bold(test.dotted_name),
                        reason,
                    )
                )

        # Captured output for non-failing tests
        if not self.args.quiet_stdout:
            failing_tests = {x[0] for x in self.all_errors}
            for test in list(self.stdout_output) + list(self.stderr_errput):
                if test not in failing_tests:
                    self.displayStdout(test)
                    self.displayStderr(test)

        # Actual tracebacks and captured output for failing tests
        for test, color_func, outcome, err in self.all_errors:
            # Header Line
            self.stream.writeln(
                f"\n{color_func(outcome)} in {self.colors.bold(test.dotted_name)}"
            )

            # Traceback
            if not self.args.no_tracebacks:
                relevant_frames = []
                for i, frame in enumerate(err.traceback_lines):
                    debug(
                        "\n"
                        f"{'*' * 30}Frame {i}:{'*' * 30}\n" + self.colors.yellow(frame),
                        level=3,
                    )
                    # Ignore useless frames
                    if self.verbose < 4:
                        if frame.strip() == "Traceback (most recent call last):":
                            continue
                    # Done with this frame, capture it.
                    relevant_frames.append(frame)
                self.stream.write("".join(relevant_frames))

            # Captured output for failing tests
            self.displayStdout(test)
            self.displayStderr(test)

    def wasSuccessful(self) -> bool:
        """
        Tells whether or not the overall run was successful.
        """
        if self.args.minimum_coverage is not None:
            if self.coverage_percent < self.args.minimum_coverage:
                self.stream.writeln(
                    self.colors.red(
                        "Coverage of {}% is below minimum level of {}%".format(
                            self.coverage_percent, self.args.minimum_coverage
                        )
                    )
                )
                return False

        # fail if no tests are run.
        if not any(
            (
                self.errors,
                self.expectedFailures,
                self.failures,
                self.passing,
                self.skipped,
                self.unexpectedSuccesses,
            )
        ):
            return False
        return not self.all_errors and not self.unexpectedSuccesses



import os
import pathlib
from os.path import dirname
import platform
import shutil
import sys
import tempfile
from textwrap import dedent
import unittest
from unittest.mock import MagicMock, patch

from green import loader
from green.loader import GreenTestLoader, flattenTestSuite


class TestToProtoTestList(unittest.TestCase):
    def test_moduleImportFailure(self):
        """
        toProtoTestList() raises import errors normally
        """
        suite = MagicMock()
        suite.__class__.__name__ = "ModuleImportFailure"
        suite.__str__.return_value = "exception_method (other_stuff)"
        suite.exception_method.side_effect = AttributeError
        self.assertRaises(AttributeError, loader.toProtoTestList, (suite,))

    def test_moduleImportFailureIgnored(self):
        """
        toProtoTestList() does not raise errors when doing completions
        """
        suite = MagicMock()
        suite.__class__.__name__ = "ModuleImportFailure"
        suite.__str__.return_value = "exception_method other_stuff"
        suite.exception_method.side_effect = AttributeError
        self.assertEqual(loader.toProtoTestList(suite, doing_completions=True), [])


class TestToParallelTargets(unittest.TestCase):
    def setUp(self):
        super().setUp()

        class FakeModule:
            pass

        class FakeModule2:
            pass

        self._fake_module_name = "my_test_module"
        self._fake_module_name2 = "my_test_module2"
        sys.modules[self._fake_module_name] = FakeModule
        sys.modules[self._fake_module_name2] = FakeModule2

    def tearDown(self):
        del sys.modules[self._fake_module_name]
        del sys.modules[self._fake_module_name2]
        super().tearDown()

    def test_methods_with_no_constraints(self):
        """
        toParallelTargets() returns only module names.
        """

        class NormalTestCase(unittest.TestCase):
            def runTest(self):
                pass

        NormalTestCase.__module__ = self._fake_module_name

        targets = loader.toParallelTargets(NormalTestCase(), [])
        self.assertEqual(targets, [self._fake_module_name])

    def test_methods_with_constraints(self):
        """
        toParallelTargets() returns test names when constrained.
        """

        class NormalTestCase(unittest.TestCase):
            def runTest(self):
                pass

        NormalTestCase.__module__ = self._fake_module_name
        full_name = "my_test_module.NormalTestCase.runTest"

        # Pass full_name as a single string to validate that it is handled correctly.
        targets = loader.toParallelTargets(NormalTestCase(), full_name)
        self.assertEqual(targets, [full_name])

    def test_filter_out_dot(self):
        """
        toParallelTargets() correctly returns modules when '.' is in target list
        """

        class NormalTestCase(unittest.TestCase):
            def runTest(self):
                pass

        class NormalTestCase2(unittest.TestCase):
            def runTest(self):
                pass

        NormalTestCase.__module__ = self._fake_module_name
        NormalTestCase2.__module__ = self._fake_module_name2

        targets = loader.toParallelTargets([NormalTestCase(), NormalTestCase2()], ["."])
        self.assertEqual(targets, ["my_test_module", "my_test_module2"])

    def test_ignore_doctest(self):
        """
        toParallelTargets() ignores"""


class TestCompletions(unittest.TestCase):
    def test_completionBad(self):
        """
        Bad match generates no completions
        """
        self.assertEqual("", loader.getCompletions("garbage.in"))

    def test_completionExact(self):
        """
        Correct completions are generated for an exact match.
        """
        c = set(loader.getCompletions("green").split("\n"))
        self.assertIn("green", c)
        self.assertIn("green.test", c)
        self.assertIn("green.test.test_loader", c)
        self.assertIn("green.test.test_loader.TestCompletions", c)
        self.assertIn("green.test.test_loader.TestCompletions.test_completionExact", c)

    def test_completionPartialShort(self):
        """
        Correct completions generated for short partial match.
        """
        cwd = os.getcwd()
        green_parent = dirname(dirname(dirname(os.path.abspath(__file__))))
        os.chdir(green_parent)
        self.addCleanup(os.chdir, cwd)
        c = set(loader.getCompletions("gre").split("\n"))
        self.assertIn("green", c)
        self.assertIn("green.test", c)
        self.assertIn("green.test.test_loader", c)
        self.assertIn("green.test.test_loader.TestCompletions", c)
        self.assertIn(
            "green.test.test_loader.TestCompletions.test_completionPartialShort", c
        )

    def test_completionPartial(self):
        """
        Correct completions generated for partial match.  2nd target ignored.
        """
        c = set(loader.getCompletions(["green.te", "green"]).split("\n"))
        self.assertIn("green.test", c)
        self.assertIn("green.test.test_loader", c)
        self.assertIn("green.test.test_loader.TestCompletions", c)
        self.assertIn(
            "green.test.test_loader.TestCompletions.test_completionPartial", c
        )
        self.assertNotIn("green", c)

    def test_completionEmpty(self):
        """
        An empty target generates completions for the whole directory
        """
        cwd = os.getcwd()
        tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tmpdir)
        os.chdir(tmpdir)
        self.addCleanup(os.chdir, cwd)
        os.mkdir("the_pkg")
        fh = open(os.path.join("the_pkg", "__init__.py"), "w")
        fh.write("")
        fh.close()
        fh = open(os.path.join("the_pkg", "test_things.py"), "w")
        fh.write(
            dedent(
                """
            import unittest

            class A(unittest.TestCase):
                def testOne(self):
                    pass
                def testTwo(self):
                    pass
            """
            )
        )
        fh.close()
        c = set(loader.getCompletions("").split("\n"))
        self.assertIn("the_pkg", c)
        self.assertIn("the_pkg.test_things", c)
        self.assertIn("the_pkg.test_things.A.testOne", c)
        self.assertIn("the_pkg.test_things.A.testTwo", c)

    def test_completionDot(self):
        """
        A '.' target generates completions for the whole directory
        """
        cwd = os.getcwd()
        tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tmpdir)
        os.chdir(tmpdir)
        self.addCleanup(os.chdir, cwd)
        os.mkdir("my_pkg")
        fh = open(os.path.join("my_pkg", "__init__.py"), "w")
        fh.write("")
        fh.close()
        fh = open(os.path.join("my_pkg", "test_things.py"), "w")
        fh.write(
            dedent(
                """
            import unittest

            class A(unittest.TestCase):
                def testOne(self):
                    pass
                def testTwo(self):
                    pass
            """
            )
        )
        fh.close()
        c = set(loader.getCompletions(".").split("\n"))
        self.assertIn("my_pkg", c)
        self.assertIn("my_pkg.test_things", c)
        self.assertIn("my_pkg.test_things.A.testOne", c)
        self.assertIn("my_pkg.test_things.A.testTwo", c)

    def test_completionIgnoresErrors(self):
        """
        Errors in one module don't block the remaining completions
        """
        cwd = os.getcwd()
        tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tmpdir)
        os.chdir(tmpdir)
        self.addCleanup(os.chdir, cwd)
        os.mkdir("my_pkg2")
        fh = open(os.path.join("my_pkg2", "__init__.py"), "w")
        fh.write("")
        fh.close()
        fh = open(os.path.join("my_pkg2", "test_crash01.py"), "w")
        contents = dedent(
            """
            import unittest

            class A(unittest.TestCase):
                def testOne(self):
                    pass
                def testTwo(self):
                    pass
            """
        )
        fh.write(contents)
        fh.close()
        fh = open(os.path.join("my_pkg2", "test_crash02.py"), "w")
        fh.write("import moocow")
        fh.close()
        fh = open(os.path.join("my_pkg2", "test_crash03.py"), "w")
        fh.write(contents)
        fh.close()
        c = set(loader.getCompletions(".").split("\n"))
        self.assertIn("my_pkg2", c)
        self.assertIn("my_pkg2.test_crash01", c)
        self.assertIn("my_pkg2.test_crash01.A.testOne", c)
        self.assertIn("my_pkg2.test_crash01.A.testTwo", c)
        self.assertIn("my_pkg2.test_crash03", c)
        self.assertIn("my_pkg2.test_crash03.A.testOne", c)
        self.assertIn("my_pkg2.test_crash03.A.testTwo", c)


class TestIsPackage(unittest.TestCase):

    def setUp(self):
        tmpdir = tempfile.mkdtemp()
        self.tmp_path = pathlib.Path(tmpdir)
        self.addCleanup(shutil.rmtree, tmpdir)

    def test_yes(self):
        """
        A package is identified.
        """
        (self.tmp_path / "__init__.py").write_text("pass\n", encoding="utf-8")
        self.assertTrue(loader.isPackage(self.tmp_path))

    def test_no(self):
        """
        A non-package is identified
        """
        self.assertFalse(loader.isPackage(self.tmp_path))


class TestDottedModule(unittest.TestCase):
    def test_bad_path(self):
        """
        A bad path causes an exception
        """
        self.assertRaises(
            ValueError, loader.findDottedModuleAndParentDir, tempfile.tempdir
        )

    def test_good_path(self):
        """
        A good path gets (dotted_module, parent) properly returned
        """
        tmpdir = tempfile.mkdtemp()
        os.makedirs(os.path.join(tmpdir, "a", "b", "c", "d"))
        package_init = os.path.join(tmpdir, "a", "b", "c", "__init__.py")
        subpkg_init = os.path.join(tmpdir, "a", "b", "c", "d", "__init__.py")
        module_name = "stuff.py"
        module = os.path.join(tmpdir, "a", "b", "c", "d", module_name)
        for filename in [package_init, subpkg_init, module]:
            fh = open(filename, "w")
            fh.write("pass\n")
            fh.close()
        self.assertEqual(
            loader.findDottedModuleAndParentDir(module),
            ("c.d.stuff", os.path.join(tmpdir, "a", "b")),
        )


class TestLoadTestsFromTestCase(unittest.TestCase):
    def setUp(self):
        self.loader = GreenTestLoader()

    def test_runTest(self):
        """
        When a testcase has no matching method names, but does have a runTest,
        use that instead.
        """

        class MyTestCase(unittest.TestCase):
            def helper1(self):
                pass

            def helper2(self):
                pass

            def runTest(self):
                pass

        suite = self.loader.loadTestsFromTestCase(MyTestCase)
        self.assertEqual(suite.countTestCases(), 1)
        self.assertEqual(suite._tests[0]._testMethodName, "runTest")

    def test_normal(self):
        """
        Normal test methods get loaded
        """

        class Normal(unittest.TestCase):
            def test_method1(self):
                pass

            def test_method2(self):
                pass

        suite = self.loader.loadTestsFromTestCase(Normal)
        self.assertEqual(suite.countTestCases(), 2)
        self.assertEqual(
            {x._testMethodName for x in suite._tests},
            {"test_method1", "test_method2"},
        )

    def test_isTestCaseDisabled(self):
        """
        TestCases disabled by nose generators don't get loaded
        """

        class HasDisabled(unittest.TestCase):
            def test_method(self):
                pass

            test_method.__test__ = False

        suite = self.loader.loadTestsFromTestCase(HasDisabled)
        self.assertEqual(suite.countTestCases(), 0)


class TestLoadFromModuleFilename(unittest.TestCase):
    def setUp(self):
        self.loader = GreenTestLoader()

    def test_skipped_module(self):
        """
        A module that wants to be skipped gets skipped
        """
        tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tmpdir)
        filename = os.path.join(tmpdir, "skipped_module.py")
        fh = open(filename, "w")
        fh.write(
            dedent(
                """
            import unittest
            raise unittest.case.SkipTest
            class NotReached(unittest.TestCase):
                def test_one(self):
                    pass
                def test_two(self):
                    pass
            """
            )
        )
        fh.close()
        suite = self.loader.loadFromModuleFilename(filename)
        self.assertEqual(suite.countTestCases(), 1)
        self.assertRaises(
            unittest.case.SkipTest,
            getattr(suite._tests[0], suite._tests[0]._testMethodName),
        )


class TestDiscover(unittest.TestCase):
    def setUp(self):
        self.loader = GreenTestLoader()

    @patch("green.loader.os.path.isdir")
    @patch("green.loader.debug")
    @patch("green.loader.os.listdir")
    def test_oserror(self, mock_listdir, mock_debug, mock_isdir):
        """
        discover() prints a debug message and moves on when ecountering an OSError
        """
        mock_isdir.return_value = True
        mock_listdir.side_effect = OSError()
        tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tmpdir)
        self.loader.discover(os.path.join(tmpdir, "garbage_in"))
        self.assertEqual(len(mock_debug.mock_calls), 1)

    def test_bad_input(self):
        """
        discover() raises ImportError when passed a non-directory
        """
        tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tmpdir)
        self.assertRaises(
            ImportError, self.loader.discover, os.path.join(tmpdir, "garbage_in")
        )
        filename = os.path.join(tmpdir, "some_file.py")
        fh = open(filename, "w")
        fh.write("pass\n")
        fh.close()
        self.assertRaises(ImportError, self.loader.discover, filename)

    def test_bad_pkg_name(self):
        """
        If the directory is an invalid package name, don't bother looking in
        it.
        """
        tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tmpdir)
        startdir = os.getcwd()
        os.chdir(tmpdir)
        self.addCleanup(os.chdir, startdir)
        bad_pkg_name = "1badname"
        os.mkdir(bad_pkg_name)
        tmp_subdir = os.path.join(tmpdir, bad_pkg_name)
        fh = open(os.path.join(tmp_subdir, "__init__.py"), "w")
        fh.write("\n")
        fh.close()
        named_module = os.path.join(os.path.basename(tmp_subdir), "named_module.py")
        fh = open(named_module, "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testPass(self):
                    pass
            """
            )
        )
        fh.close()
        self.assertEqual(self.loader.discover(tmpdir), None)

    def test_symlink(self):
        """
        If the directory is a symlink, it should be skipped.
        """
        if platform.system() == "Windows":  # pragma: no cover
            self.skipTest("This test is for posix-specific behavior")
        tmpdir = tempfile.mkdtemp()
        tmpdir2 = tempfile.mkdtemp()
        os.symlink(tmpdir, os.path.join(tmpdir2, "link"))
        self.addCleanup(shutil.rmtree, tmpdir)
        startdir = os.getcwd()
        os.chdir(tmpdir)
        self.addCleanup(os.chdir, startdir)
        pkg_name = "realpkg"
        os.mkdir(pkg_name)
        tmp_subdir = os.path.join(tmpdir, pkg_name)
        fh = open(os.path.join(tmp_subdir, "__init__.py"), "w")
        fh.write("\n")
        fh.close()
        named_module = os.path.join(os.path.basename(tmp_subdir), "test_module.py")
        fh = open(named_module, "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testPass(self):
                    pass
            """
            )
        )
        fh.close()
        self.assertEqual(self.loader.discover(tmpdir2), None)


class TestLoadTargets(unittest.TestCase):
    # Setup
    @classmethod
    def setUpClass(cls):
        cls.startdir = os.getcwd()
        cls.container_dir = tempfile.mkdtemp()

    @classmethod
    def tearDownClass(cls):
        if os.getcwd() != cls.startdir:
            os.chdir(cls.startdir)
        cls.startdir = None
        shutil.rmtree(cls.container_dir)

    def setUp(self):
        os.chdir(self.container_dir)
        self.tmpdir = tempfile.mkdtemp(dir=self.container_dir)
        self.loader = GreenTestLoader()

    def tearDown(self):
        os.chdir(self.container_dir)
        shutil.rmtree(self.tmpdir)

    # Tests
    def test_returnIsLoadable(self):
        """
        Results returned by toParallelTargets should be loadable by
        loadTargets(), even if they aren't directly loadable through a package
        relative to the current working directory.
        """
        tests_dir = tempfile.mkdtemp(dir=self.tmpdir)
        # No __init__.py in the directory!
        fh = open(os.path.join(tests_dir, "test_not_in_pkg.py"), "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testPass(self):
                    pass
            """
            )
        )
        fh.close()
        # Discover stuff
        suite = self.loader.loadTargets(".")
        # This should resolve it to the module that's not importable from here
        test = loader.toParallelTargets(suite, [])[0]
        self.loader.loadTargets(test)

    def test_emptyDirAbsolute(self):
        """
        Absolute path to empty directory returns None
        """
        tests = self.loader.loadTargets(self.tmpdir)
        self.assertTrue(tests is None)

    def test_emptyDirRelative(self):
        """
        Relative path to empty directory returns None
        """
        os.chdir(self.tmpdir)
        os.chdir("..")
        tests = self.loader.loadTargets(os.path.dirname(self.tmpdir))
        self.assertEqual(tests, None)

    def test_emptyDirDot(self):
        """
        '.' while in an empty directory returns None
        """
        os.chdir(self.tmpdir)
        tests = self.loader.loadTargets(".")
        self.assertTrue(tests is None)

    def test_relativeDotDir(self):
        """
        Dotted relative path to empty directory returns None
        """
        os.chdir(self.tmpdir)
        os.chdir("..")
        target = os.path.join(".", os.path.basename(self.tmpdir))
        tests = self.loader.loadTargets(target)
        self.assertTrue(tests is None)

    def test_BigDirWithAbsoluteImports(self):
        """
        Big dir discovers tests and doesn't crash on absolute import
        """
        sub_tmpdir = tempfile.mkdtemp(dir=self.tmpdir)
        pkg_name = os.path.basename(sub_tmpdir)
        # Child setup
        # pkg/__init__.py
        fh = open(os.path.join(sub_tmpdir, "__init__.py"), "w")
        fh.write("\n")
        fh.close()
        # pkg/target_module.py
        fh = open(os.path.join(sub_tmpdir, "target_module.py"), "w")
        fh.write("a = 1\n")
        fh.close()
        # pkg/test/__init__.py
        os.mkdir(os.path.join(sub_tmpdir, "test"))
        fh = open(os.path.join(sub_tmpdir, "test", "__init__.py"), "w")
        fh.write("\n")
        fh.close()
        # pkg/test/test_target_module.py
        fh = open(os.path.join(sub_tmpdir, "test", "test_target_module.py"), "w")
        fh.write(
            dedent(
                """
            import unittest
            import {}.target_module
            class A(unittest.TestCase):
                def testPass(self):
                    pass
            """.format(
                    pkg_name
                )
            )
        )
        fh.close()
        # Load the tests
        os.chdir(self.tmpdir)
        test_suite = self.loader.loadTargets(pkg_name)
        self.assertEqual(test_suite.countTestCases(), 1)
        # Dotted name should start with the package!
        self.assertEqual(
            pkg_name + ".test.test_target_module.A.testPass",
            loader.toProtoTestList(test_suite)[0].dotted_name,
        )

    def test_DirWithInit(self):
        """
        Dir empty other than blank __init__.py returns None
        """
        # Parent directory setup
        os.chdir(self.tmpdir)
        os.chdir("..")
        # Child setup
        target = os.path.join(self.tmpdir, "__init__.py")
        fh = open(target, "w")
        fh.write("\n")
        fh.close()
        fh = open(os.path.join(self.tmpdir, "test_module_with_init.py"), "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testPass(self):
                    pass
            """
            )
        )
        fh.close()
        # Load the tests
        module_name = os.path.basename(self.tmpdir)
        tests = self.loader.loadTargets(module_name)
        self.assertEqual(tests.countTestCases(), 1)

    def test_DottedName(self):
        """
        Importing a module via dotted name loads the tests.
        """
        # Parent directory setup
        os.chdir(self.tmpdir)
        sub_tmpdir = tempfile.mkdtemp(dir=self.tmpdir)
        basename = os.path.basename(sub_tmpdir)
        # Child setup
        fh = open(os.path.join(basename, "__init__.py"), "w")
        fh.write("\n")
        fh.close()
        fh = open(os.path.join(basename, "test_module_dotted_name.py"), "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testPass(self):
                    pass
            """
            )
        )
        fh.close()
        # Load the tests
        module_name = basename + ".test_module_dotted_name"
        tests = self.loader.loadTargets(module_name)
        self.assertEqual(tests.countTestCases(), 1)

    def test_DottedNamePackageFromPath(self):
        """
        Importing a package from path loads the tests.
        """
        # Child setup

        tmp_subdir = tempfile.mkdtemp(dir=self.tmpdir)
        fh = open(os.path.join(tmp_subdir, "__init__.py"), "w")
        fh.write("\n")
        fh.close()
        fh = open(os.path.join(tmp_subdir, "test_module.py"), "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testPass(self):
                    pass
            """
            )
        )
        fh.close()
        # Go somewhere else, but setup the path
        os.chdir(self.startdir)
        sys.path.insert(0, self.tmpdir)
        # Load the tests
        tests = self.loader.loadTargets(os.path.basename(tmp_subdir))
        sys.path.remove(self.tmpdir)
        self.assertTrue(tests.countTestCases(), 1)

    def test_ModuleByName(self):
        """
        A module in a package can be loaded by filename.
        """
        os.chdir(self.tmpdir)
        tmp_subdir = tempfile.mkdtemp(dir=self.tmpdir)
        fh = open(os.path.join(tmp_subdir, "__init__.py"), "w")
        fh.write("\n")
        fh.close()
        named_module = os.path.join(os.path.basename(tmp_subdir), "named_module.py")
        fh = open(named_module, "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testPass(self):
                    pass
            """
            )
        )
        fh.close()
        # Load the tests
        tests = self.loader.loadTargets(named_module)
        try:
            self.assertEqual(tests.countTestCases(), 1)
        except:
            raise
        finally:
            shutil.rmtree(tmp_subdir)

    def test_MalformedModuleByName(self):
        """
        Importing malformed module by name creates test that raises
        ImportError.
        """
        fh = open(os.path.join(self.tmpdir, "__init__.py"), "w")
        fh.write("\n")
        fh.close()
        malformed_module = os.path.join(
            os.path.basename(self.tmpdir), "malformed_module.py"
        )
        fh = open(malformed_module, "w")
        fh.write("This is a malformed module.")
        fh.close()
        # Load the tests
        tests = self.loader.loadTargets(malformed_module)
        self.assertEqual(tests.countTestCases(), 1)
        test = tests._tests[0]
        test_method = getattr(test, test._testMethodName)
        self.assertRaises(ImportError, test_method)

    def test_partiallyGoodName(self):
        """
        Don't crash loading module.object with existing module but not object.
        """
        # Parent directory setup
        os.chdir(self.tmpdir)
        sub_tmpdir = tempfile.mkdtemp(dir=self.tmpdir)
        basename = os.path.basename(sub_tmpdir)
        # Child setup
        fh = open(os.path.join(basename, "__init__.py"), "w")
        fh.write("\n")
        fh.close()
        fh = open(os.path.join(basename, "existing_module.py"), "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testPass(self):
                    pass
            """
            )
        )
        fh.close()
        # Load the tests
        module_name = basename + ".existing_module.nonexistant_object"
        tests = self.loader.loadTargets(module_name)
        self.assertEqual(tests, None)

    def test_multiple_targets(self):
        """
        Specifying multiple targets causes them all to be tested.
        """
        sub_tmpdir = tempfile.mkdtemp(dir=self.tmpdir)
        # pkg/__init__.py
        fh = open(os.path.join(sub_tmpdir, "__init__.py"), "w")
        fh.write("\n")
        fh.close()
        # pkg/test/test_target1.py
        fh = open(os.path.join(sub_tmpdir, "test_target1.py"), "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testPasses(self):
                    pass
            """
            )
        )
        fh.close()
        # pkg/test/test_target2.py
        fh = open(os.path.join(sub_tmpdir, "test_target2.py"), "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testPasses(self):
                    pass
            """
            )
        )
        fh.close()
        # Load the tests
        os.chdir(self.tmpdir)
        pkg = os.path.basename(sub_tmpdir)
        tests = self.loader.loadTargets(
            [pkg + "." + "test_target1", pkg + "." + "test_target2"]
        )
        self.assertEqual(tests.countTestCases(), 2)

    def test_duplicate_targets(self):
        """
        Specifying duplicate targets does not cause duplicate loading.
        """
        sub_tmpdir = tempfile.mkdtemp(dir=self.tmpdir)
        fh = open(os.path.join(sub_tmpdir, "__init__.py"), "w")
        fh.write("\n")
        fh.close()
        fh = open(os.path.join(sub_tmpdir, "test_dupe_target.py"), "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testPasses(self):
                    pass
            """
            )
        )
        fh.close()

        os.chdir(self.tmpdir)
        pkg = os.path.basename(sub_tmpdir)
        tests = self.loader.loadTargets(
            [
                pkg + "." + "test_dupe_target",
                pkg + "." + "test_dupe_target",
                pkg + "." + "test_dupe_target",
            ]
        )
        self.assertEqual(tests.countTestCases(), 1)

    def test_syntax_error_by_filename(self):
        """
        Loading a module by file name with a syntax error produces a crash.
        """
        sub_tmpdir = tempfile.mkdtemp(dir=self.tmpdir)
        fh = open(os.path.join(sub_tmpdir, "mod_with_import_error.py"), "w")
        fh.write("this is a syntax error")
        fh.close()

        os.chdir(sub_tmpdir)
        hit_exception = False
        try:
            self.loader.loadTargets("mod_with_import_error.py")
        except Exception as e:
            self.assertIn("invalid syntax", str(e))
            self.assertIn("mod_with_import_error.py", str(e))
            hit_exception = True
        if not hit_exception:
            self.fail("An exception should have been raised. :-(")

    def test_syntax_error_by_dotname(self):
        """
        Loading a module by dotname with a syntax error produces a crash.
        """
        sub_tmpdir = tempfile.mkdtemp(dir=self.tmpdir)
        fh = open(os.path.join(sub_tmpdir, "mod_with_syntax_error.py"), "w")
        fh.write(
            dedent(
                """
            import unittest
            class TestSyntax(unittest.TestCase):
                def test_syntax(self):
                    syntax error
            """
            )
        )
        fh.close()

        os.chdir(sub_tmpdir)
        hit_exception = False
        try:
            self.loader.loadTargets("mod_with_syntax_error.TestSyntax.test_syntax")
        except Exception as e:
            self.assertIn("invalid syntax", str(e))
            self.assertIn("mod_with_syntax_error.TestSyntax.test_syntax", str(e))
            hit_exception = True
        if not hit_exception:
            self.fail(
                "An exception should have been raised about the syntax error. :-("
            )

    def test_file_pattern(self):
        """
        Specifying a file pattern causes only matching files to be loaded
        """
        sub_tmpdir = tempfile.mkdtemp(dir=self.tmpdir)
        # pkg/__init__.py
        fh = open(os.path.join(sub_tmpdir, "__init__.py"), "w")
        fh.write("\n")
        fh.close()
        # pkg/test/target1_tests.py
        fh = open(os.path.join(sub_tmpdir, "target1_tests.py"), "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testPasses(self):
                    pass
            """
            )
        )
        fh.close()
        # pkg/test/target2_tests.py
        fh = open(os.path.join(sub_tmpdir, "target2_tests.py"), "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testPasses(self):
                    pass
            """
            )
        )
        fh.close()
        # pkg/test/test_target999.py: NOT a match.
        fh = open(os.path.join(sub_tmpdir, "test_target999.py"), "w")
        fh.write(
            dedent(
                """
            import unittest
            class A(unittest.TestCase):
                def testPasses(self):
                    pass
            """
            )
        )
        fh.close()
        # Load the tests
        os.chdir(self.tmpdir)
        pkg = os.path.basename(sub_tmpdir)
        tests = self.loader.loadTargets(pkg, file_pattern="*_tests.py")
        self.assertEqual(tests.countTestCases(), 2)


class TestFlattenTestSuite(unittest.TestCase):
    # Setup
    @classmethod
    def setUpClass(cls):
        cls.startdir = os.getcwd()
        cls.container_dir = tempfile.mkdtemp()

    @classmethod
    def tearDownClass(cls):
        if os.getcwd() != cls.startdir:
            os.chdir(cls.startdir)
        cls.startdir = None
        shutil.rmtree(cls.container_dir)

    def setUp(self):
        os.chdir(self.container_dir)
        self.tmpdir = tempfile.mkdtemp(dir=self.container_dir)
        self.loader = GreenTestLoader()

    def tearDown(self):
        os.chdir(self.container_dir)
        shutil.rmtree(self.tmpdir)

    @patch("green.loader.GreenTestLoader.suiteClass")
    @patch("green.loader.DocTestSuite")
    def test_docTests(self, mock_doc_test_suite, mock_suite_class):
        """
        flattenTestSuite injects the test module name into the doctest's .__module__
        """
        mock_test = MagicMock()
        mock_iter = MagicMock(return_value=iter([mock_test]))
        mock_suite = MagicMock()
        mock_suite.__iter__ = mock_iter
        mock_doc_test_suite.return_value = mock_suite
        module = MagicMock()
        test_module_name = "test.module"
        module.__name__ = test_module_name
        module.doctest_modules = ["real.module"]
        flattenTestSuite((), module)

        self.assertEqual(mock_test.__module__, test_module_name)



from argparse import Namespace
from argparse import ArgumentParser
from io import StringIO
import sys
import unittest
from unittest.mock import MagicMock, patch

from green import djangorunner
from green.config import mergeConfig


class TestDjangoMissing(unittest.TestCase):
    def test_importError(self):
        """
        Raises ImportError if Django is not available
        """
        self.assertRaises(ImportError, djangorunner.django_missing)


class TestDjangoRunner(unittest.TestCase):
    def setUp(self):
        try:
            djangorunner.DjangoRunner()
        except ImportError:
            raise unittest.SkipTest("Django is not installed")
        saved_stdout = sys.stdout
        self.stream = StringIO()
        sys.stdout = self.stream
        self.addCleanup(setattr, sys, "stdout", saved_stdout)

    def test_run_testsWithLabel(self):
        """
        Labeled tests run okay
        """
        dr = djangorunner.DjangoRunner()
        dr.setup_test_environment = MagicMock()
        dr.setup_databases = MagicMock()
        dr.teardown_databases = MagicMock()
        dr.teardown_test_environment = MagicMock()

        dr.run_tests(("green.test.test_version",), testing=True)

        self.assertIn("OK", self.stream.getvalue())

    def test_run_testsWithoutLabel(self):
        """
        Not passing in a label causes the targets to be ['.']
        """
        dr = djangorunner.DjangoRunner()
        dr.setup_test_environment = MagicMock()
        dr.setup_databases = MagicMock()
        dr.teardown_databases = MagicMock()
        dr.teardown_test_environment = MagicMock()

        with patch.object(dr.loader, "loadTargets") as mock_loadTargets:
            dr.run_tests((), testing=True)

        mock_loadTargets.assert_called_with(["."])
        self.assertIn("No Tests Found", self.stream.getvalue())

    def test_run_testsWithBadInput(self):
        """
        Bad input causes a ValueError to be raised
        """
        dr = djangorunner.DjangoRunner()
        dr.setup_test_environment = MagicMock()
        dr.setup_databases = MagicMock()

        self.assertRaises(ValueError, dr.run_tests, None, True)

    @patch("green.djangorunner.GreenTestSuite")
    @patch("green.djangorunner.run")
    def test_run_noTests(self, mock_run, mock_GreenTestSuite):
        """
        If no tests are found, we create an empty test suite and run it.
        """
        dr = djangorunner.DjangoRunner()

        dr.setup_test_environment = MagicMock()
        dr.setup_databases = MagicMock()
        dr.teardown_databases = MagicMock()
        dr.teardown_test_environment = MagicMock()

        mock_GreenTestSuite.return_value = 123

        with patch.object(dr.loader, "loadTargets", return_value=None):
            dr.run_tests((), testing=True)

        self.assertEqual(mock_run.call_args[0][0], 123)

    @patch("green.djangorunner.mergeConfig")
    @patch("green.djangorunner.GreenTestSuite")
    @patch("green.djangorunner.run")
    def test_run_coverage(self, mock_run, mock_GreenTestSuite, mock_mergeConfig):
        """
        If no tests are found, we create an empty test suite and run it.
        """
        args = mergeConfig(Namespace())
        args.run_coverage = True
        args.cov = MagicMock()
        mock_mergeConfig.return_value = args
        dr = djangorunner.DjangoRunner()

        dr.setup_test_environment = MagicMock()
        dr.setup_databases = MagicMock()
        dr.teardown_databases = MagicMock()
        dr.teardown_test_environment = MagicMock()

        mock_GreenTestSuite.return_value = 123

        with patch.object(dr.loader, "loadTargets", return_value=None):
            dr.run_tests((), testing=True)

        self.assertEqual(mock_run.call_args[0][0], 123)

    def test_check_verbosity_argument_recognised(self):
        """
        Ensure that the python manage.py test command
        recognises the --green-verbosity flag
        """
        dr = djangorunner.DjangoRunner()
        dr.setup_test_environment = MagicMock()
        dr.setup_databases = MagicMock()
        dr.teardown_databases = MagicMock()
        dr.teardown_test_environment = MagicMock()
        from django.core.management.commands.test import Command as TestCommand

        test_command = TestCommand()
        test_command.test_runner = "green.djangorunner.DjangoRunner"
        parser = ArgumentParser()
        test_command.add_arguments(parser)
        args = parser.parse_args()
        self.assertIn("verbose", args)

    def test_check_default_verbosity(self):
        """
        If no verbosity is passed, default value is set
        """
        dr = djangorunner.DjangoRunner()
        dr.setup_test_environment = MagicMock()
        dr.setup_databases = MagicMock()
        dr.teardown_databases = MagicMock()
        dr.teardown_test_environment = MagicMock()
        from django.core.management.commands.test import Command as TestCommand

        test_command = TestCommand()
        test_command.test_runner = "green.djangorunner.DjangoRunner"
        parser = ArgumentParser()
        test_command.add_arguments(parser)
        args = parser.parse_args()
        self.assertEqual(args.verbose, -1)

    def test_run_with_verbosity_flag(self):
        """
        Tests should run fine if verbosity is passed
        through CLI flag
        """
        dr = djangorunner.DjangoRunner()
        dr.setup_test_environment = MagicMock()
        dr.setup_databases = MagicMock()
        dr.teardown_databases = MagicMock()
        dr.teardown_test_environment = MagicMock()
        dr.verbose = 2
        saved_loadTargets = dr.loader.loadTargets
        dr.loader.loadTargets = MagicMock()
        self.addCleanup(setattr, dr.loader, "loadTargets", saved_loadTargets)
        self.assertEqual((dr.run_tests((), testing=True)), 0)



"""
To try running Django tests using green you can run:

    ./manage.py test --testrunner=green.djangorunner.DjangoRunner

To make the change permanent for your project, in settings.py add:

    TEST_RUNNER="green.djangorunner.DjangoRunner"
"""

from __future__ import annotations

from argparse import ArgumentParser, Namespace
import pathlib
import os
import sys
from typing import Any, Final, Sequence

from green.config import mergeConfig
from green.loader import GreenTestLoader
from green.output import GreenStream
from green.runner import run
from green.suite import GreenTestSuite

# If we're not being run from an actual django project, set up django config
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "green.djangorunner")
BASE_DIR = pathlib.Path(__file__).absolute().parent.parent
SECRET_KEY: Final[str] = ")9^_e(=cisybdt4m4+fs+_wb%d$!9mpcoy0um^alvx%gexj#jv"
DEBUG: bool = True
TEMPLATE_DEBUG: bool = True
ALLOWED_HOSTS: Sequence[str] = []
INSTALLED_APPS: Final[Sequence[str]] = (
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "myapp",
)
MIDDLEWARE_CLASSES: Final[Sequence[str]] = (
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.auth.middleware.SessionAuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
)
ROOT_URLCONF: Final[str] = "myproj.urls"
WSGI_APPLICATION: Final[str] = "myproj.wsgi.application"
DATABASES: Final[dict[str, dict[str, str]]] = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": str(BASE_DIR / "db.sqlite3"),
    }
}
LANGUAGE_CODE: Final[str] = "en-us"
TIME_ZONE: Final[str] = "UTC"
USE_I18N: bool = True
USE_L10N: bool = True
USE_TZ: bool = True
STATIC_URL: Final[str] = "/static/"
# End of django fake config stuff


def django_missing() -> None:
    raise ImportError("No django module installed")


try:
    import django

    if django.VERSION[:2] < (1, 6):  # pragma: no cover
        raise ImportError("Green integration supports Django 1.6+")
    from django.test.runner import DiscoverRunner

    class DjangoRunner(DiscoverRunner):
        def __init__(self, verbose: int = -1, **kwargs: Any):
            super().__init__(**kwargs)
            self.verbose = verbose
            self.loader = GreenTestLoader()

        @classmethod
        def add_arguments(cls, parser: ArgumentParser) -> None:
            parser.add_argument(
                "--green-verbosity",
                action="store",
                dest="verbose",
                default=-1,
                type=int,
                help="""
                    Green 'verbose' level for tests.  Value should be an integer
                    that green supports.  For example: --green-verbosity 3""",
            )
            super().add_arguments(parser)

        # FIXME: extra_tests is not used, we should either use it or update the
        #  documentation accordingly.
        def run_tests(
            self,
            test_labels: list[str] | tuple[str, ...],
            extra_tests: Any = None,
            **kwargs: Any,
        ):
            """
            Run the unit tests for all the test labels in the provided list.

            Test labels should be dotted Python paths to test modules, test
            classes, or test methods.

            A list of 'extra' tests may also be provided; these tests
            will be added to the test suite.

            Returns the number of tests that failed.
            """
            # Django setup
            self.setup_test_environment()
            django_db = self.setup_databases()

            # Green
            if isinstance(test_labels, tuple):
                test_labels = list(test_labels)
            else:
                raise ValueError("test_labels should be a tuple of strings")
            if not test_labels:
                test_labels = ["."]

            args = mergeConfig(Namespace())
            if self.verbose != -1:
                args.verbose = self.verbose
            args.targets = test_labels
            stream = GreenStream(sys.stdout)
            suite = self.loader.loadTargets(args.targets)
            if not suite:
                suite = GreenTestSuite()
            result = run(suite, stream, args)

            # Django teardown
            self.teardown_databases(django_db)
            self.teardown_test_environment()
            return self.suite_result(suite, result)

except ImportError:  # pragma: no cover
    DjangoRunner = django_missing  # type: ignore



from io import StringIO
import platform
import unittest
from unittest.mock import MagicMock, patch

from green.output import Colors, GreenStream, debug
import green.output


class TestDebug(unittest.TestCase):
    def testDebug(self):
        """
        debug() works as we expect
        """
        orig_logging = green.output.logging.debug
        s = StringIO()
        green.output.logging.debug = s.write
        green.output.debug_level = 0
        debug("Nothing should happen", level=1)
        self.assertEqual("", s.getvalue())
        green.output.debug_level = 2
        debug("Something should happen", level=1)
        self.assertNotEqual("", s.getvalue())

        green.output.logging.debug = orig_logging


class TestColors(unittest.TestCase):
    def testTermcolorTrue(self):
        """
        termcolor=True results in terminal output
        """
        c = Colors(termcolor=True)
        self.assertTrue(c.termcolor)
        self.assertTrue(len(c.bold("")) > 0)

    def testTermcolorFalse(self):
        """
        termcolor=False results in no terminal output
        """
        c = Colors(termcolor=False)
        self.assertFalse(c.termcolor)
        self.assertFalse(len(c.bold("")) > 0)

    def testTermcolorAuto(self):
        """
        termcolor=None causes termcolor autodetected and set to True or False
        """
        c = Colors()
        self.assertTrue(c.termcolor in [True, False])

    def testUp(self):
        """
        calling up gives us a non-blank string
        """
        c = Colors()
        up = c.up()
        self.assertEqual(type(up), str)
        self.assertNotEqual(up, "")

    def testTerminalColorsDoNotCrash(self):
        """
        terminal colors don't crash, and they output something
        """
        c = Colors(termcolor=True)
        for func in [
            c.bold,
            c.blue,
            c.green,
            c.red,
            c.yellow,
            c.passing,
            c.failing,
            c.error,
            c.skipped,
            c.unexpectedSuccess,
            c.expectedFailure,
            c.moduleName,
        ]:
            self.assertTrue(len(func("")) > 0)
        # c.className is a special case
        c.className("")


class TestGreenStream(unittest.TestCase):
    def testFormatText(self):
        """
        formatText returns the input text by default
        """
        s = StringIO()
        gs = GreenStream(s)
        msg = "Unindented line.\n  Indented.\n    Double-indented.\n\n\n"
        self.assertEqual(gs.formatText(msg), str(msg))

    def testBadStringType(self):
        """
        passing the wrong stream type to GreenStream gets auto-converted
        """
        s = StringIO()
        gs = GreenStream(s)
        msg = "some string"
        gs.write(bytes(msg, "utf-8"))
        self.assertEqual(s.getvalue(), msg)

    def testDisableWindowsTrue(self):
        """
        disable_windows=True: ANSI color codes are present in the stream
        """
        c = Colors(termcolor=True)
        s = StringIO()
        gs = GreenStream(s, disable_windows=True)
        msg = c.red("some colored string")
        gs.write(msg)
        self.assertEqual(len(gs.stream.getvalue()), len(msg))

    @unittest.skipIf(
        platform.system() != "Windows",
        "Colorama won't strip ANSI unless running on Windows",
    )
    def testDisableWindowsFalse(self):
        """
        disable_windows=False: Colorama strips ANSI color codes from the stream
        """
        c = Colors(termcolor=True)
        s = StringIO()
        gs = GreenStream(s, override_appveyor=True, disable_windows=False)
        colored_msg = c.red("a")
        gs.write(colored_msg)
        import colorama

        self.assertTrue(issubclass(type(gs.stream), colorama.ansitowin32.StreamWrapper))

    @patch("green.output.unidecode")
    def testUnidecodeAppveyor(self, mock_unidecode):
        """
        When I'm on Appveyor, I run text through Unidecode
        """
        mock_unidecode.return_value = "something"
        s = StringIO()
        gs = GreenStream(s, override_appveyor=True)
        gs.write("something")
        self.assertTrue(mock_unidecode.called)

    @patch("green.output.unidecode")
    def testUnidecodeDisabled(self, mock_unidecode):
        """
        Unidecode can be disabled
        """
        mock_unidecode.return_value = "something"
        s = StringIO()
        gs = GreenStream(s, override_appveyor=True, disable_unidecode=True)
        gs.write("something")
        self.assertFalse(mock_unidecode.called)

    def testWritelines(self):
        """
        Compatibility function writelines(lines) repeatedly calls write()
        """
        s = StringIO()
        gs = GreenStream(s)
        gs.write = MagicMock()
        gs.writelines(["one", "two", "three"])
        self.assertEqual(len(gs.write.mock_calls), 3)

    def testCoverageDetection(self):
        """
        write() detects a coverage percentage flying by
        """
        s = StringIO()
        gs = GreenStream(s)
        gs.write(
            "\n---------------------------------------------------\nTOTAL                   896    367    59%\nRan"
        )
        self.assertEqual(gs.coverage_percent, 59)

    def testEncodingMirrors(self):
        """
        The encoding of a stream gets mirrored through
        """
        s = StringIO()
        encoding = "aoeu"
        try:
            encoding = s.encoding
        except:
            s.encoding = encoding
        gs = GreenStream(s)
        self.assertEqual(gs.encoding, encoding)

    def testEncodingDefault(self):
        """
        The encoding defaults to 'UTF-8' if we can't find an encoding.
        """
        s = MagicMock(spec=1)
        gs = GreenStream(s)
        self.assertEqual(gs.encoding, "UTF-8")



"""
Handle running unittests suites in parallel.
"""

from __future__ import annotations

import logging
import multiprocessing
import multiprocessing.pool
from multiprocessing.pool import MaybeEncodingError  # type: ignore
from multiprocessing.pool import Pool
from multiprocessing import util  # type: ignore

import os
import random
import sys
import tempfile
import traceback
from typing import (
    Type,
    TYPE_CHECKING,
    Union,
    Tuple,
    Callable,
    Iterable,
    Mapping,
    Any,
    TypeVar,
)

import coverage

from green.exceptions import InitializerOrFinalizerError
from green.loader import GreenTestLoader
from green.result import proto_test, ProtoTest, ProtoTestResult

if TYPE_CHECKING:
    from types import TracebackType
    from queue import Queue

    from multiprocessing.context import SpawnContext, SpawnProcess
    from multiprocessing.pool import ApplyResult
    from multiprocessing.queues import SimpleQueue

    from green.suite import GreenTestSuite
    from green.runner import InitializerOrFinalizer
    from green.result import RunnableTestT

    ExcInfoType = Union[
        Tuple[Type[BaseException], BaseException, TracebackType],
        Tuple[None, None, None],
    ]
    _T = TypeVar("_T")


# Super-useful debug function for finding problems in the subprocesses, and it
# even works on windows
def ddebug(msg: str, err: ExcInfoType | None = None) -> None:  # pragma: no cover
    """
    err can be an instance of sys.exc_info() -- which is the latest traceback
    info
    """
    if err:
        error_string = "".join(traceback.format_exception(*err))
    else:
        error_string = ""
    sys.__stdout__.write(f"({os.getpid()}) {msg} {error_string}\n")
    sys.__stdout__.flush()


class ProcessLogger:
    """
    I am used by LoggingDaemonlessPool to get crash output out to the logger,
    instead of having process crashes be silent.
    """

    def __init__(self, callable: Callable) -> None:
        self.__callable = callable

    def __call__(self, *args, **kwargs) -> Any:
        try:
            return self.__callable(*args, **kwargs)
        except Exception:
            # Here we add some debugging help. If multiprocessing's
            # debugging is on, it will arrange to log the traceback
            logger = multiprocessing.get_logger()
            if not logger.handlers:
                logger.addHandler(logging.StreamHandler())
            logger.error(traceback.format_exc())
            logger.handlers[0].flush()
            # Re-raise the original exception so the Pool worker can
            # clean up
            raise


class LoggingDaemonlessPool(Pool):
    """
    I make a pool of workers which can get crash output to the logger, run processes not as daemons,
    and which run finalizers.
    """

    _wrap_exception: bool = True

    @staticmethod
    def Process(ctx: SpawnContext, *args: Any, **kwargs: Any) -> SpawnProcess:
        return ctx.Process(daemon=False, *args, **kwargs)

    def apply_async(
        self,
        func: Callable[[Any, Any], _T],  # should be the poolRunner method.
        args: Iterable = (),
        kwargs: Mapping[str, Any] | None = None,
        callback: Callable[[_T], Any] | None = None,
        error_callback: Callable[[BaseException], Any] | None = None,
    ) -> ApplyResult[_T]:
        if kwargs is None:
            kwargs = {}
        return super().apply_async(
            ProcessLogger(func), args, kwargs, callback, error_callback
        )

    def __init__(
        self,
        processes: int | None = None,
        initializer: Callable | None = None,
        initargs: Iterable[Any] = (),
        maxtasksperchild: int | None = None,
        context: Any | None = None,
        # Green specific:
        finalizer: Callable | None = None,
        finalargs: Iterable[Any] = (),
    ):
        self._finalizer = finalizer
        self._finalargs = finalargs
        super().__init__(processes, initializer, initargs, maxtasksperchild, context)

    def _repopulate_pool(self):
        return self._repopulate_pool_static(
            self._ctx,
            self.Process,
            self._processes,
            self._pool,
            self._inqueue,
            self._outqueue,
            self._initializer,
            self._initargs,
            self._maxtasksperchild,
            self._wrap_exception,
            self._finalizer,
            self._finalargs,
        )

    @staticmethod
    def _repopulate_pool_static(
        ctx: SpawnContext,
        Process: Callable,  # LoggingDaemonlessPool.Process
        processes: int,
        pool: list[Callable],  # list of LoggingDaemonlessPool.Process
        inqueue: SimpleQueue,
        outqueue: SimpleQueue,
        initializer: InitializerOrFinalizer,
        initargs: tuple,
        maxtasksperchild: int | None,
        wrap_exception: bool,
        finalizer: InitializerOrFinalizer,
        finalargs: tuple,
    ) -> None:
        """
        Bring the number of pool processes up to the specified number,
        for use after reaping workers which have exited.
        """
        for i in range(processes - len(pool)):
            w = Process(
                ctx,
                target=worker,
                args=(
                    inqueue,
                    outqueue,
                    initializer,
                    initargs,
                    maxtasksperchild,
                    wrap_exception,
                    finalizer,
                    finalargs,
                ),
            )
            w.name = w.name.replace("Process", "PoolWorker")
            w.start()
            pool.append(w)
            util.debug("added worker")


def worker(
    inqueue: SimpleQueue,
    outqueue: SimpleQueue,
    initializer: InitializerOrFinalizer | None = None,
    initargs: tuple = (),
    maxtasks: int | None = None,
    wrap_exception: bool = False,
    finalizer: Callable | None = None,
    finalargs: tuple = (),
):  # pragma: no cover
    # TODO: revisit this assert; these statements are skipped by the python
    #  compiler in optimized mode.
    assert maxtasks is None or (isinstance(maxtasks, int) and maxtasks > 0)
    put = outqueue.put
    get = inqueue.get

    writer = getattr(inqueue, "_writer", None)
    if writer is not None:
        writer.close()
    reader = getattr(outqueue, "_reader", None)
    if reader is not None:
        reader.close()

    if initializer is not None:
        try:
            initializer(*initargs)
        except InitializerOrFinalizerError as e:
            print(str(e))

    completed = 0
    while maxtasks is None or (maxtasks and completed < maxtasks):
        try:
            task = get()
        except (EOFError, OSError):
            util.debug("worker got EOFError or OSError -- exiting")
            break

        if task is None:
            util.debug("worker got sentinel -- exiting")
            break

        job, i, func, args, kwds = task
        try:
            result = (True, func(*args, **kwds))
        except Exception as result_error:
            if wrap_exception:
                result_error = ExceptionWithTraceback(
                    result_error, result_error.__traceback__
                )
            result = (False, result_error)
        try:
            put((job, i, result))
        except Exception as e:
            wrapped = MaybeEncodingError(e, result[1])
            util.debug("Possible encoding error while sending result: %s" % (wrapped))
            put((job, i, (False, wrapped)))
        completed += 1

    if finalizer:
        try:
            finalizer(*finalargs)
        except InitializerOrFinalizerError as e:
            print(str(e))

    util.debug("worker exiting after %d tasks" % completed)


# Unmodified (see above)
class RemoteTraceback(Exception):  # pragma: no cover
    def __init__(self, tb: str):
        self.tb = tb

    def __str__(self) -> str:
        return self.tb


# Unmodified (see above)
class ExceptionWithTraceback(Exception):  # pragma: no cover
    def __init__(self, exc: BaseException, tb: TracebackType | None):
        tb_lines = traceback.format_exception(type(exc), exc, tb)
        tb_text = "".join(tb_lines)
        self.exc = exc
        self.tb = '\n"""\n%s"""' % tb_text

    def __reduce__(self) -> Tuple[Callable, Tuple[BaseException, str]]:
        return rebuild_exc, (self.exc, self.tb)


# Unmodified (see above)
def rebuild_exc(exc: BaseException, tb: str):  # pragma: no cover
    exc.__cause__ = RemoteTraceback(tb)
    return exc


multiprocessing.pool.worker = worker  # type: ignore
# END of Worker Finalization Monkey Patching
# -----------------------------------------------------------------------------


def poolRunner(
    target: str,
    queue: Queue,
    coverage_number: int | None = None,
    omit_patterns: str | Iterable[str] | None = None,
    cov_config_file: bool = True,
) -> None:  # pragma: no cover
    """
    I am the function that pool worker processes run.  I run one unit test.

    coverage_config_file is a special option that is either a string specifying
    the custom coverage config file or the special default value True (which
    causes coverage to search for it's standard config files).
    """
    # Each pool worker gets his own temp directory, to avoid having tests that
    # are used to taking turns using the same temp file name from interfering
    # with eachother.  So long as the test doesn't use a hard-coded temp
    # directory, anyway.
    saved_tempdir = tempfile.tempdir
    tempfile.tempdir = tempfile.mkdtemp()

    def raise_internal_failure(msg: str) -> None:
        err = sys.exc_info()
        t = ProtoTest()
        t.module = "green.loader"
        t.class_name = "N/A"
        t.description = msg
        t.method_name = "poolRunner"
        result.startTest(t)
        result.addError(t, err)
        result.stopTest(t)
        queue.put(result)
        cleanup()

    def cleanup() -> None:
        # Restore the state of the temp directory
        tempfile.tempdir = saved_tempdir
        queue.put(None)
        # Finish coverage
        if coverage_number:
            cov.stop()
            cov.save()

    # Each pool starts its own coverage, later combined by the main process.
    if coverage_number:
        cov = coverage.coverage(
            data_file=".coverage.{}_{}".format(
                coverage_number, random.randint(0, 10000)
            ),
            omit=omit_patterns,
            config_file=cov_config_file,
        )
        cov._warn_no_data = False
        cov.start()

    # What to do each time an individual test is started
    already_sent = set()

    def start_callback(test: RunnableTestT) -> None:
        # Let the main process know what test we are starting
        test_proto = proto_test(test)
        if test_proto not in already_sent:
            queue.put(test_proto)
            already_sent.add(test_proto)

    def finalize_callback(test_result: ProtoTestResult) -> None:
        # Let the main process know what happened with the test run
        queue.put(test_result)

    result = ProtoTestResult(start_callback, finalize_callback)
    test: GreenTestSuite | None
    try:
        loader = GreenTestLoader()
        test = loader.loadTargets(target)
    except:
        raise_internal_failure("Green encountered an error loading the unit test.")
        return

    if test is not None and getattr(test, "run", False):
        # Loading was successful, lets do this
        try:
            test.run(result)
            # If your class setUpClass(self) method crashes, the test doesn't
            # raise an exception, but it does add an entry to errors.  Some
            # other things add entries to errors as well, but they all call the
            # finalize callback.
            if (
                result
                and (not result.finalize_callback_called)
                and getattr(result, "errors", False)
            ):
                queue.put(test)
                queue.put(result)
        except:
            # Some frameworks like testtools record the error AND THEN let it
            # through to crash things.  So we only need to manufacture another
            # error if the underlying framework didn't, but either way we don't
            # want to crash.
            if result.errors:
                queue.put(result)
            else:
                try:
                    err = sys.exc_info()
                    result.startTest(test)
                    result.addError(test, err)
                    result.stopTest(test)
                    queue.put(result)
                except:
                    raise_internal_failure(
                        "Green encountered an error when running the test."
                    )
                    return
    else:
        # loadTargets() returned an object without a run() method, probably
        # None
        description = (
            f'Test loader returned an un-runnable object.  Is "{target}" '
            "importable from your current location?  Maybe you "
            "forgot an __init__.py in your directory?  Unrunnable "
            f"object looks like: {test} of type {type(test)} with dir {dir(test)}"
        )
        no_run_error = (TypeError, TypeError(description), None)
        t = ProtoTest()
        t.description = description
        target_list = target.split(".")
        if len(target_list) > 1:
            t.module = ".".join(target_list[:-2])
            t.class_name = target_list[-2]
            t.method_name = target_list[-1]
        else:
            t.module = target
            t.class_name = "UnknownClass"
            t.method_name = "unknown_method"
        result.startTest(t)
        # Ignoring that no_run_error traceback is None.
        result.addError(t, no_run_error)  # type: ignore[arg-type]
        result.stopTest(t)
        queue.put(result)

    cleanup()
